# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :class:`iris.common.metadata.BaseMetadata`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from collections import OrderedDict
import unittest.mock as mock
from unittest.mock import sentinel

import numpy.ma as ma
import numpy as np

from iris.common.lenient import _LENIENT, _qualname
from iris.common.metadata import BaseMetadata, CubeMetadata


class Test(tests.IrisTest):
    def setUp(self):
        self.standard_name = mock.sentinel.standard_name
        self.long_name = mock.sentinel.long_name
        self.var_name = mock.sentinel.var_name
        self.units = mock.sentinel.units
        self.attributes = mock.sentinel.attributes
        self.cls = BaseMetadata

    def test_repr(self):
        metadata = self.cls(
            standard_name=self.standard_name,
            long_name=self.long_name,
            var_name=self.var_name,
            units=self.units,
            attributes=self.attributes,
        )
        fmt = (
            "BaseMetadata(standard_name={!r}, long_name={!r}, "
            "var_name={!r}, units={!r}, attributes={!r})"
        )
        expected = fmt.format(
            self.standard_name,
            self.long_name,
            self.var_name,
            self.units,
            self.attributes,
        )
        self.assertEqual(expected, repr(metadata))

    def test__fields(self):
        expected = (
            "standard_name",
            "long_name",
            "var_name",
            "units",
            "attributes",
        )
        self.assertEqual(expected, self.cls._fields)


class Test___eq__(tests.IrisTest):
    def setUp(self):
        self.kwargs = dict(
            standard_name=sentinel.standard_name,
            long_name=sentinel.long_name,
            var_name=sentinel.var_name,
            units=sentinel.units,
            attributes=sentinel.attributes,
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(**self.kwargs)

    def test_lenient_service(self):
        qualname___eq__ = _qualname(self.cls.__eq__)
        self.assertIn(qualname___eq__, _LENIENT)
        self.assertTrue(_LENIENT[qualname___eq__])
        self.assertTrue(_LENIENT[self.cls.__eq__])

    def test_cannot_compare_non_class(self):
        result = self.metadata.__eq__(None)
        self.assertIs(NotImplemented, result)

    def test_cannot_compare_different_class(self):
        other = CubeMetadata(*(None,) * len(CubeMetadata._fields))
        result = self.metadata.__eq__(other)
        self.assertIs(NotImplemented, result)

    def test_lenient(self):
        return_value = sentinel.return_value
        with mock.patch(
            "iris.common.metadata._LENIENT", return_value=True
        ) as mlenient:
            with mock.patch.object(
                self.cls, "_compare_lenient", return_value=return_value
            ) as mcompare:
                result = self.metadata.__eq__(self.metadata)

        self.assertEqual(return_value, result)
        self.assertEqual(1, mcompare.call_count)
        (arg,), kwargs = mcompare.call_args
        self.assertEqual(id(self.metadata), id(arg))
        self.assertEqual(dict(), kwargs)

        self.assertEqual(1, mlenient.call_count)
        (arg,), kwargs = mlenient.call_args
        self.assertEqual(_qualname(self.cls.__eq__), _qualname(arg))
        self.assertEqual(dict(), kwargs)

    def test_strict_same(self):
        self.assertTrue(self.metadata.__eq__(self.metadata))
        other = self.cls(**self.kwargs)
        self.assertTrue(self.metadata.__eq__(other))
        self.assertTrue(other.__eq__(self.metadata))

    def test_strict_different(self):
        self.kwargs["var_name"] = None
        other = self.cls(**self.kwargs)
        self.assertFalse(self.metadata.__eq__(other))
        self.assertFalse(other.__eq__(self.metadata))


class Test___lt__(tests.IrisTest):
    def setUp(self):
        self.cls = BaseMetadata
        self.one = self.cls(1, 1, 1, 1, 1)
        self.two = self.cls(1, 1, 1, 2, 1)
        self.none = self.cls(1, 1, 1, None, 1)
        self.attributes = self.cls(1, 1, 1, 1, 10)

    def test__ascending_lt(self):
        result = self.one < self.two
        self.assertTrue(result)

    def test__descending_lt(self):
        result = self.two < self.one
        self.assertFalse(result)

    def test__none_rhs_operand(self):
        result = self.one < self.none
        self.assertFalse(result)

    def test__none_lhs_operand(self):
        result = self.none < self.one
        self.assertTrue(result)

    def test__ignore_attributes(self):
        result = self.one < self.attributes
        self.assertFalse(result)
        result = self.attributes < self.one
        self.assertFalse(result)


class Test___ne__(tests.IrisTest):
    def setUp(self):
        self.cls = BaseMetadata
        self.metadata = self.cls(*(None,) * len(self.cls._fields))
        self.other = sentinel.other

    def test_notimplemented(self):
        return_value = NotImplemented
        with mock.patch.object(
            self.cls, "__eq__", return_value=return_value
        ) as mocker:
            result = self.metadata.__ne__(self.other)

        self.assertIs(return_value, result)
        self.assertEqual(1, mocker.call_count)
        (arg,), kwargs = mocker.call_args
        self.assertEqual(self.other, arg)
        self.assertEqual(dict(), kwargs)

    def test_negate_true(self):
        return_value = True
        with mock.patch.object(
            self.cls, "__eq__", return_value=return_value
        ) as mocker:
            result = self.metadata.__ne__(self.other)

        self.assertFalse(result)
        (arg,), kwargs = mocker.call_args
        self.assertEqual(self.other, arg)
        self.assertEqual(dict(), kwargs)

    def test_negate_false(self):
        return_value = False
        with mock.patch.object(
            self.cls, "__eq__", return_value=return_value
        ) as mocker:
            result = self.metadata.__ne__(self.other)

        self.assertTrue(result)
        (arg,), kwargs = mocker.call_args
        self.assertEqual(self.other, arg)
        self.assertEqual(dict(), kwargs)


class Test__combine(tests.IrisTest):
    def setUp(self):
        self.kwargs = dict(
            standard_name="standard_name",
            long_name="long_name",
            var_name="var_name",
            units="units",
            attributes=dict(one=sentinel.one, two=sentinel.two),
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(**self.kwargs)

    def test_lenient(self):
        return_value = sentinel._combine_lenient
        other = sentinel.other
        with mock.patch(
            "iris.common.metadata._LENIENT", return_value=True
        ) as mlenient:
            with mock.patch.object(
                self.cls, "_combine_lenient", return_value=return_value
            ) as mcombine:
                result = self.metadata._combine(other)

                self.assertEqual(1, mlenient.call_count)
                (arg,), kwargs = mlenient.call_args
                self.assertEqual(self.metadata.combine, arg)
                self.assertEqual(dict(), kwargs)

                self.assertEqual(return_value, result)
                self.assertEqual(1, mcombine.call_count)
                (arg,), kwargs = mcombine.call_args
                self.assertEqual(other, arg)
                self.assertEqual(dict(), kwargs)

    def test_strict(self):
        dummy = sentinel.dummy
        values = self.kwargs.copy()
        values["standard_name"] = dummy
        values["var_name"] = dummy
        values["attributes"] = dummy
        other = self.cls(**values)
        with mock.patch("iris.common.metadata._LENIENT", return_value=False):
            result = self.metadata._combine(other)

        expected = [
            None if values[field] == dummy else values[field]
            for field in self.cls._fields
        ]
        self.assertEqual(expected, result)


class Test__combine_lenient(tests.IrisTest):
    def setUp(self):
        self.cls = BaseMetadata
        self.none = self.cls(*(None,) * len(self.cls._fields))._asdict()
        self.names = dict(
            standard_name=sentinel.standard_name,
            long_name=sentinel.long_name,
            var_name=sentinel.var_name,
        )

    def test_strict_units(self):
        left = self.none.copy()
        left["units"] = "K"
        right = left.copy()
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        expected = list(left.values())
        self.assertEqual(expected, lmetadata._combine_lenient(rmetadata))
        self.assertEqual(expected, rmetadata._combine_lenient(lmetadata))

    def test_strict_units_different(self):
        left = self.none.copy()
        right = self.none.copy()
        left["units"] = "K"
        right["units"] = "km"
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        result = lmetadata._combine_lenient(rmetadata)
        expected = list(self.none.values())
        self.assertEqual(expected, result)
        result = rmetadata._combine_lenient(lmetadata)
        self.assertEqual(expected, result)

    def test_strict_units_different_none(self):
        left = self.none.copy()
        right = self.none.copy()
        left["units"] = "K"
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        result = lmetadata._combine_lenient(rmetadata)
        expected = list(self.none.values())
        self.assertEqual(expected, result)

        result = rmetadata._combine_lenient(lmetadata)
        self.assertEqual(expected, result)

    def test_attributes(self):
        left = self.none.copy()
        right = self.none.copy()
        ldict = dict(item=sentinel.left)
        rdict = dict(item=sentinel.right)
        left["attributes"] = ldict
        right["attributes"] = rdict
        rmetadata = self.cls(**right)
        return_value = sentinel.return_value
        with mock.patch.object(
            self.cls,
            "_combine_lenient_attributes",
            return_value=return_value,
        ) as mocker:
            lmetadata = self.cls(**left)
            result = lmetadata._combine_lenient(rmetadata)

        expected = self.none.copy()
        expected["attributes"] = return_value
        expected = list(expected.values())
        self.assertEqual(expected, result)

        self.assertEqual(1, mocker.call_count)
        args, kwargs = mocker.call_args
        expected = (ldict, rdict)
        self.assertEqual(expected, args)
        self.assertEqual(dict(), kwargs)

    def test_attributes_non_mapping_different(self):
        left = self.none.copy()
        right = self.none.copy()
        ldict = dict(item=sentinel.left)
        rdict = sentinel.right
        left["attributes"] = ldict
        right["attributes"] = rdict
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        expected = list(self.none.copy().values())
        self.assertEqual(expected, lmetadata._combine_lenient(rmetadata))
        self.assertEqual(expected, rmetadata._combine_lenient(lmetadata))

    def test_attributes_non_mapping_different_none(self):
        left = self.none.copy()
        right = self.none.copy()
        ldict = dict(item=sentinel.left)
        left["attributes"] = ldict
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        result = lmetadata._combine_lenient(rmetadata)
        expected = self.none.copy()
        expected["attributes"] = ldict
        expected = list(expected.values())
        self.assertEqual(expected, result)

        result = rmetadata._combine_lenient(lmetadata)
        self.assertEqual(expected, result)

    def test_names(self):
        left = self.none.copy()
        left.update(self.names)
        right = left.copy()
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        expected = list(left.values())
        self.assertEqual(expected, lmetadata._combine_lenient(rmetadata))
        self.assertEqual(expected, rmetadata._combine_lenient(lmetadata))

    def test_names_different(self):
        dummy = sentinel.dummy
        left = self.none.copy()
        right = self.none.copy()
        left.update(self.names)
        right["standard_name"] = dummy
        right["long_name"] = dummy
        right["var_name"] = dummy
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        expected = list(self.none.copy().values())
        self.assertEqual(expected, lmetadata._combine_lenient(rmetadata))
        self.assertEqual(expected, rmetadata._combine_lenient(lmetadata))

    def test_names_different_none(self):
        left = self.none.copy()
        right = self.none.copy()
        left.update(self.names)
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        result = lmetadata._combine_lenient(rmetadata)
        expected = list(left.values())
        self.assertEqual(expected, result)

        result = rmetadata._combine_lenient(lmetadata)
        self.assertEqual(expected, result)


class Test__combine_lenient_attributes(tests.IrisTest):
    def setUp(self):
        self.values = OrderedDict(
            one="one",
            two="two",
            three=np.int16(123),
            four=np.arange(10),
            five=ma.arange(10),
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(*(None,) * len(self.cls._fields))
        self.dummy = sentinel.dummy

    def test_same(self):
        left = self.values.copy()
        right = self.values.copy()

        result = self.metadata._combine_lenient_attributes(left, right)
        expected = left
        self.assertDictEqual(expected, result)

        result = self.metadata._combine_lenient_attributes(right, left)
        self.assertDictEqual(expected, result)

    def test_different(self):
        left = self.values.copy()
        right = self.values.copy()
        left["two"] = left["four"] = self.dummy

        result = self.metadata._combine_lenient_attributes(left, right)
        expected = self.values.copy()
        for key in ["two", "four"]:
            del expected[key]
        self.assertDictEqual(expected, result)

        result = self.metadata._combine_lenient_attributes(right, left)
        self.assertDictEqual(expected, result)

    def test_different_none(self):
        left = self.values.copy()
        right = self.values.copy()
        left["one"] = left["three"] = left["five"] = None

        result = self.metadata._combine_lenient_attributes(left, right)
        expected = self.values.copy()
        for key in ["one", "three", "five"]:
            del expected[key]
        self.assertDictEqual(expected, result)

        result = self.metadata._combine_lenient_attributes(right, left)
        self.assertDictEqual(expected, result)

    def test_extra(self):
        left = self.values.copy()
        right = self.values.copy()
        left["extra_left"] = "extra_left"
        right["extra_right"] = "extra_right"

        result = self.metadata._combine_lenient_attributes(left, right)
        expected = self.values.copy()
        expected["extra_left"] = left["extra_left"]
        expected["extra_right"] = right["extra_right"]
        self.assertDictEqual(expected, result)

        result = self.metadata._combine_lenient_attributes(right, left)
        self.assertDictEqual(expected, result)


class Test__combine_strict_attributes(tests.IrisTest):
    def setUp(self):
        self.values = OrderedDict(
            one="one",
            two="two",
            three=np.int32(123),
            four=np.arange(10),
            five=ma.arange(10),
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(*(None,) * len(self.cls._fields))
        self.dummy = sentinel.dummy

    def test_same(self):
        left = self.values.copy()
        right = self.values.copy()

        result = self.metadata._combine_strict_attributes(left, right)
        expected = left
        self.assertDictEqual(expected, result)

        result = self.metadata._combine_strict_attributes(right, left)
        self.assertDictEqual(expected, result)

    def test_different(self):
        left = self.values.copy()
        right = self.values.copy()
        left["one"] = left["three"] = self.dummy

        result = self.metadata._combine_strict_attributes(left, right)
        expected = self.values.copy()
        for key in ["one", "three"]:
            del expected[key]
        self.assertDictEqual(expected, result)

        result = self.metadata._combine_strict_attributes(right, left)
        self.assertDictEqual(expected, result)

    def test_different_none(self):
        left = self.values.copy()
        right = self.values.copy()
        left["one"] = left["three"] = left["five"] = None

        result = self.metadata._combine_strict_attributes(left, right)
        expected = self.values.copy()
        for key in ["one", "three", "five"]:
            del expected[key]
        self.assertDictEqual(expected, result)

        result = self.metadata._combine_strict_attributes(right, left)
        self.assertDictEqual(expected, result)

    def test_extra(self):
        left = self.values.copy()
        right = self.values.copy()
        left["extra_left"] = "extra_left"
        right["extra_right"] = "extra_right"

        result = self.metadata._combine_strict_attributes(left, right)
        expected = self.values.copy()
        self.assertDictEqual(expected, result)

        result = self.metadata._combine_strict_attributes(right, left)
        self.assertDictEqual(expected, result)


class Test__compare_lenient(tests.IrisTest):
    def setUp(self):
        self.cls = BaseMetadata
        self.none = self.cls(*(None,) * len(self.cls._fields))._asdict()
        self.names = dict(
            standard_name=sentinel.standard_name,
            long_name=sentinel.long_name,
            var_name=sentinel.var_name,
        )

    def test_name_same(self):
        left = self.none.copy()
        left.update(self.names)
        right = left.copy()
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        with mock.patch.object(
            self.cls, "_is_attributes", return_value=False
        ) as mocker:
            self.assertTrue(lmetadata._compare_lenient(rmetadata))
            self.assertTrue(rmetadata._compare_lenient(lmetadata))

        # mocker not called for "units" nor "var_name" members.
        expected = (len(self.cls._fields) - 2) * 2
        self.assertEqual(expected, mocker.call_count)

    def test_name_same_lenient_false__long_name_different(self):
        left = self.none.copy()
        left.update(self.names)
        right = left.copy()
        right["long_name"] = sentinel.dummy
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        with mock.patch.object(
            self.cls, "_is_attributes", return_value=False
        ) as mocker:
            self.assertFalse(lmetadata._compare_lenient(rmetadata))
            self.assertFalse(rmetadata._compare_lenient(lmetadata))

        # mocker not called for "units" nor "var_name" members.
        expected = (len(self.cls._fields) - 2) * 2
        self.assertEqual(expected, mocker.call_count)

    def test_name_same_lenient_true__var_name_different(self):
        left = self.none.copy()
        left.update(self.names)
        right = left.copy()
        right["var_name"] = sentinel.dummy
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        with mock.patch.object(
            self.cls, "_is_attributes", return_value=False
        ) as mocker:
            self.assertTrue(lmetadata._compare_lenient(rmetadata))
            self.assertTrue(rmetadata._compare_lenient(lmetadata))

        # mocker not called for "units" nor "var_name" members.
        expected = (len(self.cls._fields) - 2) * 2
        self.assertEqual(expected, mocker.call_count)

    def test_name_different(self):
        left = self.none.copy()
        left.update(self.names)
        right = left.copy()
        right["standard_name"] = None
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        with mock.patch.object(self.cls, "_is_attributes") as mocker:
            self.assertFalse(lmetadata._compare_lenient(rmetadata))
            self.assertFalse(rmetadata._compare_lenient(lmetadata))

        self.assertEqual(0, mocker.call_count)

    def test_strict_units(self):
        left = self.none.copy()
        left.update(self.names)
        left["units"] = "K"
        right = left.copy()
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        with mock.patch.object(
            self.cls, "_is_attributes", return_value=False
        ) as mocker:
            self.assertTrue(lmetadata._compare_lenient(rmetadata))
            self.assertTrue(rmetadata._compare_lenient(lmetadata))

        # mocker not called for "units" nor "var_name" members.
        expected = (len(self.cls._fields) - 2) * 2
        self.assertEqual(expected, mocker.call_count)

    def test_strict_units_different(self):
        left = self.none.copy()
        left.update(self.names)
        left["units"] = "K"
        right = left.copy()
        right["units"] = "m"
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        with mock.patch.object(
            self.cls, "_is_attributes", return_value=False
        ) as mocker:
            self.assertFalse(lmetadata._compare_lenient(rmetadata))
            self.assertFalse(rmetadata._compare_lenient(lmetadata))

        # mocker not called for "units" nor "var_name" members.
        expected = (len(self.cls._fields) - 2) * 2
        self.assertEqual(expected, mocker.call_count)

    def test_attributes(self):
        left = self.none.copy()
        left.update(self.names)
        right = left.copy()
        ldict = dict(item=sentinel.left)
        rdict = dict(item=sentinel.right)
        left["attributes"] = ldict
        right["attributes"] = rdict
        rmetadata = self.cls(**right)
        with mock.patch.object(
            self.cls,
            "_compare_lenient_attributes",
            return_value=True,
        ) as mocker:
            lmetadata = self.cls(**left)
            self.assertTrue(lmetadata._compare_lenient(rmetadata))
            self.assertTrue(rmetadata._compare_lenient(lmetadata))

        self.assertEqual(2, mocker.call_count)
        expected = [((ldict, rdict),), ((rdict, ldict),)]
        self.assertEqual(expected, mocker.call_args_list)

    def test_attributes_non_mapping_different(self):
        left = self.none.copy()
        left.update(self.names)
        right = left.copy()
        ldict = dict(item=sentinel.left)
        rdict = sentinel.right
        left["attributes"] = ldict
        right["attributes"] = rdict
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        self.assertFalse(lmetadata._compare_lenient(rmetadata))
        self.assertFalse(rmetadata._compare_lenient(lmetadata))

    def test_attributes_non_mapping_different_none(self):
        left = self.none.copy()
        left.update(self.names)
        right = left.copy()
        ldict = dict(item=sentinel.left)
        left["attributes"] = ldict
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        self.assertTrue(lmetadata._compare_lenient(rmetadata))
        self.assertTrue(rmetadata._compare_lenient(lmetadata))

    def test_names(self):
        left = self.none.copy()
        left.update(self.names)
        left["long_name"] = None
        right = self.none.copy()
        right["long_name"] = left["standard_name"]
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        self.assertTrue(lmetadata._compare_lenient(rmetadata))
        self.assertTrue(rmetadata._combine_lenient(lmetadata))


class Test__compare_lenient_attributes(tests.IrisTest):
    def setUp(self):
        self.values = OrderedDict(
            one=sentinel.one,
            two=sentinel.two,
            three=np.int16(123),
            four=np.arange(10),
            five=ma.arange(5),
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(*(None,) * len(self.cls._fields))
        self.dummy = sentinel.dummy

    def test_same(self):
        left = self.values.copy()
        right = self.values.copy()

        self.assertTrue(self.metadata._compare_lenient_attributes(left, right))
        self.assertTrue(self.metadata._compare_lenient_attributes(right, left))

    def test_different(self):
        left = self.values.copy()
        right = self.values.copy()
        left["two"] = left["four"] = self.dummy

        self.assertFalse(
            self.metadata._compare_lenient_attributes(left, right)
        )
        self.assertFalse(
            self.metadata._compare_lenient_attributes(right, left)
        )

    def test_different_none(self):
        left = self.values.copy()
        right = self.values.copy()
        left["one"] = left["three"] = left["five"] = None

        self.assertFalse(
            self.metadata._compare_lenient_attributes(left, right)
        )
        self.assertFalse(
            self.metadata._compare_lenient_attributes(right, left)
        )

    def test_extra(self):
        left = self.values.copy()
        right = self.values.copy()
        left["extra_left"] = sentinel.extra_left
        right["extra_right"] = sentinel.extra_right

        self.assertTrue(self.metadata._compare_lenient_attributes(left, right))
        self.assertTrue(self.metadata._compare_lenient_attributes(right, left))


class Test__compare_strict_attributes(tests.IrisTest):
    def setUp(self):
        self.values = OrderedDict(
            one=sentinel.one,
            two=sentinel.two,
            three=np.int16(123),
            four=np.arange(10),
            five=ma.arange(5),
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(*(None,) * len(self.cls._fields))
        self.dummy = sentinel.dummy

    def test_same(self):
        left = self.values.copy()
        right = self.values.copy()

        self.assertTrue(self.metadata._compare_strict_attributes(left, right))
        self.assertTrue(self.metadata._compare_strict_attributes(right, left))

    def test_different(self):
        left = self.values.copy()
        right = self.values.copy()
        left["two"] = left["four"] = self.dummy

        self.assertFalse(self.metadata._compare_strict_attributes(left, right))
        self.assertFalse(self.metadata._compare_strict_attributes(right, left))

    def test_different_none(self):
        left = self.values.copy()
        right = self.values.copy()
        left["one"] = left["three"] = left["five"] = None

        self.assertFalse(self.metadata._compare_strict_attributes(left, right))
        self.assertFalse(self.metadata._compare_strict_attributes(right, left))

    def test_extra(self):
        left = self.values.copy()
        right = self.values.copy()
        left["extra_left"] = sentinel.extra_left
        right["extra_right"] = sentinel.extra_right

        self.assertFalse(self.metadata._compare_strict_attributes(left, right))
        self.assertFalse(self.metadata._compare_strict_attributes(right, left))


class Test__difference(tests.IrisTest):
    def setUp(self):
        self.kwargs = dict(
            standard_name="standard_name",
            long_name="long_name",
            var_name="var_name",
            units="units",
            attributes=dict(one=sentinel.one, two=sentinel.two),
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(**self.kwargs)

    def test_lenient(self):
        return_value = sentinel._difference_lenient
        other = sentinel.other
        with mock.patch(
            "iris.common.metadata._LENIENT", return_value=True
        ) as mlenient:
            with mock.patch.object(
                self.cls, "_difference_lenient", return_value=return_value
            ) as mdifference:
                result = self.metadata._difference(other)

                self.assertEqual(1, mlenient.call_count)
                (arg,), kwargs = mlenient.call_args
                self.assertEqual(self.metadata.difference, arg)
                self.assertEqual(dict(), kwargs)

                self.assertEqual(return_value, result)
                self.assertEqual(1, mdifference.call_count)
                (arg,), kwargs = mdifference.call_args
                self.assertEqual(other, arg)
                self.assertEqual(dict(), kwargs)

    def test_strict(self):
        dummy = sentinel.dummy
        values = self.kwargs.copy()
        values["long_name"] = dummy
        values["units"] = dummy
        other = self.cls(**values)
        method = "_difference_strict_attributes"
        with mock.patch("iris.common.metadata._LENIENT", return_value=False):
            with mock.patch.object(
                self.cls, method, return_value=None
            ) as mdifference:
                result = self.metadata._difference(other)

        expected = [
            (self.kwargs[field], dummy) if values[field] == dummy else None
            for field in self.cls._fields
        ]
        self.assertEqual(expected, result)
        self.assertEqual(1, mdifference.call_count)
        args, kwargs = mdifference.call_args
        expected = (self.kwargs["attributes"], values["attributes"])
        self.assertEqual(expected, args)
        self.assertEqual(dict(), kwargs)

        with mock.patch.object(
            self.cls, method, return_value=None
        ) as mdifference:
            result = other._difference(self.metadata)

        expected = [
            (dummy, self.kwargs[field]) if values[field] == dummy else None
            for field in self.cls._fields
        ]
        self.assertEqual(expected, result)
        self.assertEqual(1, mdifference.call_count)
        args, kwargs = mdifference.call_args
        expected = (self.kwargs["attributes"], values["attributes"])
        self.assertEqual(expected, args)
        self.assertEqual(dict(), kwargs)


class Test__difference_lenient(tests.IrisTest):
    def setUp(self):
        self.cls = BaseMetadata
        self.none = self.cls(*(None,) * len(self.cls._fields))._asdict()
        self.names = dict(
            standard_name=sentinel.standard_name,
            long_name=sentinel.long_name,
            var_name=sentinel.var_name,
        )

    def test_strict_units(self):
        left = self.none.copy()
        left["units"] = "km"
        right = left.copy()
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)
        expected = list(self.none.values())
        self.assertEqual(expected, lmetadata._difference_lenient(rmetadata))
        self.assertEqual(expected, rmetadata._difference_lenient(lmetadata))

    def test_strict_units_different(self):
        left = self.none.copy()
        right = self.none.copy()
        lunits, runits = "m", "km"
        left["units"] = lunits
        right["units"] = runits
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        result = lmetadata._difference_lenient(rmetadata)
        expected = self.none.copy()
        expected["units"] = (lunits, runits)
        expected = list(expected.values())
        self.assertEqual(expected, result)

        result = rmetadata._difference_lenient(lmetadata)
        expected = self.none.copy()
        expected["units"] = (runits, lunits)
        expected = list(expected.values())
        self.assertEqual(expected, result)

    def test_strict_units_different_none(self):
        left = self.none.copy()
        right = self.none.copy()
        lunits, runits = "m", None
        left["units"] = lunits
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        result = lmetadata._difference_lenient(rmetadata)
        expected = self.none.copy()
        expected["units"] = (lunits, runits)
        expected = list(expected.values())

        self.assertEqual(expected, result)
        result = rmetadata._difference_lenient(lmetadata)
        expected = self.none.copy()
        expected["units"] = (runits, lunits)
        expected = list(expected.values())
        self.assertEqual(expected, result)

    def test_attributes(self):
        left = self.none.copy()
        right = self.none.copy()
        ldict = dict(item=sentinel.left)
        rdict = dict(item=sentinel.right)
        left["attributes"] = ldict
        right["attributes"] = rdict
        rmetadata = self.cls(**right)
        return_value = sentinel.return_value
        with mock.patch.object(
            self.cls,
            "_difference_lenient_attributes",
            return_value=return_value,
        ) as mocker:
            lmetadata = self.cls(**left)
            result = lmetadata._difference_lenient(rmetadata)

        expected = self.none.copy()
        expected["attributes"] = return_value
        expected = list(expected.values())
        self.assertEqual(expected, result)

        self.assertEqual(1, mocker.call_count)
        args, kwargs = mocker.call_args
        expected = (ldict, rdict)
        self.assertEqual(expected, args)
        self.assertEqual(dict(), kwargs)

    def test_attributes_non_mapping_different(self):
        left = self.none.copy()
        right = self.none.copy()
        ldict = dict(item=sentinel.left)
        rdict = sentinel.right
        left["attributes"] = ldict
        right["attributes"] = rdict
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        result = lmetadata._difference_lenient(rmetadata)
        expected = self.none.copy()
        expected["attributes"] = (ldict, rdict)
        expected = list(expected.values())
        self.assertEqual(expected, result)

        result = rmetadata._difference_lenient(lmetadata)
        expected = self.none.copy()
        expected["attributes"] = (rdict, ldict)
        expected = list(expected.values())
        self.assertEqual(expected, result)

    def test_attributes_non_mapping_different_none(self):
        left = self.none.copy()
        right = self.none.copy()
        ldict = dict(item=sentinel.left)
        left["attributes"] = ldict
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        result = lmetadata._difference_lenient(rmetadata)
        expected = list(self.none.copy().values())
        self.assertEqual(expected, result)

        result = rmetadata._difference_lenient(lmetadata)
        self.assertEqual(expected, result)

    def test_names(self):
        left = self.none.copy()
        left.update(self.names)
        right = left.copy()
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        expected = list(self.none.values())
        self.assertEqual(expected, lmetadata._difference_lenient(rmetadata))
        self.assertEqual(expected, rmetadata._difference_lenient(lmetadata))

    def test_names_different(self):
        dummy = sentinel.dummy
        left = self.none.copy()
        right = self.none.copy()
        left.update(self.names)
        right["standard_name"] = dummy
        right["long_name"] = dummy
        right["var_name"] = dummy
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        result = lmetadata._difference_lenient(rmetadata)
        expected = self.none.copy()
        expected["standard_name"] = (
            left["standard_name"],
            right["standard_name"],
        )
        expected["long_name"] = (left["long_name"], right["long_name"])
        expected["var_name"] = (left["var_name"], right["var_name"])
        expected = list(expected.values())
        self.assertEqual(expected, result)

        result = rmetadata._difference_lenient(lmetadata)
        expected = self.none.copy()
        expected["standard_name"] = (
            right["standard_name"],
            left["standard_name"],
        )
        expected["long_name"] = (right["long_name"], left["long_name"])
        expected["var_name"] = (right["var_name"], left["var_name"])
        expected = list(expected.values())
        self.assertEqual(expected, result)

    def test_names_different_none(self):
        left = self.none.copy()
        right = self.none.copy()
        left.update(self.names)
        lmetadata = self.cls(**left)
        rmetadata = self.cls(**right)

        result = lmetadata._difference_lenient(rmetadata)
        expected = list(self.none.values())
        self.assertEqual(expected, result)

        result = rmetadata._difference_lenient(lmetadata)
        self.assertEqual(expected, result)


class Test__difference_lenient_attributes(tests.IrisTest):
    def setUp(self):
        self.values = OrderedDict(
            one=sentinel.one,
            two=sentinel.two,
            three=np.float(3.14),
            four=np.arange(10, dtype=np.float),
            five=ma.arange(10, dtype=np.int16),
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(*(None,) * len(self.cls._fields))
        self.dummy = sentinel.dummy

    def test_same(self):
        left = self.values.copy()
        right = self.values.copy()

        result = self.metadata._difference_lenient_attributes(left, right)
        self.assertIsNone(result)

        result = self.metadata._difference_lenient_attributes(right, left)
        self.assertIsNone(result)

    def test_different(self):
        left = self.values.copy()
        right = self.values.copy()
        left["two"] = left["four"] = self.dummy

        result = self.metadata._difference_lenient_attributes(left, right)
        for key in ["one", "three", "five"]:
            del left[key]
            del right[key]
        expected_left, expected_right = (left, right)
        result_left, result_right = result
        self.assertDictEqual(expected_left, result_left)
        self.assertDictEqual(expected_right, result_right)

        result = self.metadata._difference_lenient_attributes(right, left)
        result_left, result_right = result
        self.assertDictEqual(expected_right, result_left)
        self.assertDictEqual(expected_left, result_right)

    def test_different_none(self):
        left = self.values.copy()
        right = self.values.copy()
        left["one"] = left["three"] = left["five"] = None

        result = self.metadata._difference_lenient_attributes(left, right)
        for key in ["two", "four"]:
            del left[key]
            del right[key]
        expected_left, expected_right = (left, right)
        result_left, result_right = result
        self.assertDictEqual(expected_left, result_left)
        self.assertDictEqual(expected_right, result_right)

        result = self.metadata._difference_lenient_attributes(right, left)
        result_left, result_right = result
        self.assertDictEqual(expected_right, result_left)
        self.assertDictEqual(expected_left, result_right)

    def test_extra(self):
        left = self.values.copy()
        right = self.values.copy()
        left["extra_left"] = sentinel.extra_left
        right["extra_right"] = sentinel.extra_right
        result = self.metadata._difference_lenient_attributes(left, right)
        self.assertIsNone(result)

        result = self.metadata._difference_lenient_attributes(right, left)
        self.assertIsNone(result)


class Test__difference_strict_attributes(tests.IrisTest):
    def setUp(self):
        self.values = OrderedDict(
            one=sentinel.one,
            two=sentinel.two,
            three=np.int32(123),
            four=np.arange(10),
            five=ma.arange(10),
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(*(None,) * len(self.cls._fields))
        self.dummy = sentinel.dummy

    def test_same(self):
        left = self.values.copy()
        right = self.values.copy()

        result = self.metadata._difference_strict_attributes(left, right)
        self.assertIsNone(result)
        result = self.metadata._difference_strict_attributes(right, left)
        self.assertIsNone(result)

    def test_different(self):
        left = self.values.copy()
        right = self.values.copy()
        left["one"] = left["three"] = left["five"] = self.dummy

        result = self.metadata._difference_strict_attributes(left, right)
        expected_left = left.copy()
        expected_right = right.copy()
        for key in ["two", "four"]:
            del expected_left[key]
            del expected_right[key]
        result_left, result_right = result
        self.assertDictEqual(expected_left, result_left)
        self.assertDictEqual(expected_right, result_right)

        result = self.metadata._difference_strict_attributes(right, left)
        result_left, result_right = result
        self.assertDictEqual(expected_right, result_left)
        self.assertDictEqual(expected_left, result_right)

    def test_different_none(self):
        left = self.values.copy()
        right = self.values.copy()
        left["one"] = left["three"] = left["five"] = None

        result = self.metadata._difference_strict_attributes(left, right)
        expected_left = left.copy()
        expected_right = right.copy()
        for key in ["two", "four"]:
            del expected_left[key]
            del expected_right[key]
        result_left, result_right = result
        self.assertDictEqual(expected_left, result_left)
        self.assertDictEqual(expected_right, result_right)

        result = self.metadata._difference_strict_attributes(right, left)
        result_left, result_right = result
        self.assertDictEqual(expected_right, result_left)
        self.assertDictEqual(expected_left, result_right)

    def test_extra(self):
        left = self.values.copy()
        right = self.values.copy()
        left["extra_left"] = sentinel.extra_left
        right["extra_right"] = sentinel.extra_right

        result = self.metadata._difference_strict_attributes(left, right)
        expected_left = dict(extra_left=left["extra_left"])
        expected_right = dict(extra_right=right["extra_right"])
        result_left, result_right = result
        self.assertDictEqual(expected_left, result_left)
        self.assertDictEqual(expected_right, result_right)

        result = self.metadata._difference_strict_attributes(right, left)
        result_left, result_right = result
        self.assertDictEqual(expected_right, result_left)
        self.assertDictEqual(expected_left, result_right)


class Test__is_attributes(tests.IrisTest):
    def setUp(self):
        self.cls = BaseMetadata
        self.metadata = self.cls(*(None,) * len(self.cls._fields))
        self.field = "attributes"

    def test_field(self):
        self.assertTrue(self.metadata._is_attributes(self.field, {}, {}))

    def test_field_not_attributes(self):
        self.assertFalse(self.metadata._is_attributes(None, {}, {}))

    def test_left_not_mapping(self):
        self.assertFalse(self.metadata._is_attributes(self.field, None, {}))

    def test_right_not_mapping(self):
        self.assertFalse(self.metadata._is_attributes(self.field, {}, None))


class Test_combine(tests.IrisTest):
    def setUp(self):
        kwargs = dict(
            standard_name="standard_name",
            long_name="long_name",
            var_name="var_name",
            units="units",
            attributes="attributes",
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(**kwargs)
        self.mock_kwargs = OrderedDict(
            standard_name=sentinel.standard_name,
            long_name=sentinel.long_name,
            var_name=sentinel.var_name,
            units=sentinel.units,
            attributes=sentinel.attributes,
        )

    def test_lenient_service(self):
        qualname_combine = _qualname(self.cls.combine)
        self.assertIn(qualname_combine, _LENIENT)
        self.assertTrue(_LENIENT[qualname_combine])
        self.assertTrue(_LENIENT[self.cls.combine])

    def test_cannot_combine_non_class(self):
        emsg = "Cannot combine"
        with self.assertRaisesRegex(TypeError, emsg):
            self.metadata.combine(None)

    def test_cannot_combine_different_class(self):
        other = CubeMetadata(*(None,) * len(CubeMetadata._fields))
        emsg = "Cannot combine"
        with self.assertRaisesRegex(TypeError, emsg):
            self.metadata.combine(other)

    def test_lenient_default(self):
        return_value = self.mock_kwargs.values()
        with mock.patch.object(
            self.cls, "_combine", return_value=return_value
        ) as mocker:
            result = self.metadata.combine(self.metadata)

        self.assertEqual(self.mock_kwargs, result._asdict())
        self.assertEqual(1, mocker.call_count)
        (arg,), kwargs = mocker.call_args
        self.assertEqual(id(self.metadata), id(arg))
        self.assertEqual(dict(), kwargs)

    def test_lenient_true(self):
        return_value = self.mock_kwargs.values()
        with mock.patch.object(
            self.cls, "_combine", return_value=return_value
        ) as mcombine:
            with mock.patch.object(_LENIENT, "context") as mcontext:
                result = self.metadata.combine(self.metadata, lenient=True)

        self.assertEqual(1, mcontext.call_count)
        (arg,), kwargs = mcontext.call_args
        self.assertEqual(_qualname(self.cls.combine), arg)
        self.assertEqual(dict(), kwargs)

        self.assertEqual(result._asdict(), self.mock_kwargs)
        self.assertEqual(1, mcombine.call_count)
        (arg,), kwargs = mcombine.call_args
        self.assertEqual(id(self.metadata), id(arg))
        self.assertEqual(dict(), kwargs)

    def test_lenient_false(self):
        return_value = self.mock_kwargs.values()
        with mock.patch.object(
            self.cls, "_combine", return_value=return_value
        ) as mcombine:
            with mock.patch.object(_LENIENT, "context") as mcontext:
                result = self.metadata.combine(self.metadata, lenient=False)

        self.assertEqual(1, mcontext.call_count)
        args, kwargs = mcontext.call_args
        self.assertEqual((), args)
        self.assertEqual({_qualname(self.cls.combine): False}, kwargs)

        self.assertEqual(self.mock_kwargs, result._asdict())
        self.assertEqual(1, mcombine.call_count)
        (arg,), kwargs = mcombine.call_args
        self.assertEqual(id(self.metadata), id(arg))
        self.assertEqual(dict(), kwargs)


class Test_difference(tests.IrisTest):
    def setUp(self):
        kwargs = dict(
            standard_name="standard_name",
            long_name="long_name",
            var_name="var_name",
            units="units",
            attributes="attributes",
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(**kwargs)
        self.mock_kwargs = OrderedDict(
            standard_name=sentinel.standard_name,
            long_name=sentinel.long_name,
            var_name=sentinel.var_name,
            units=sentinel.units,
            attributes=sentinel.attributes,
        )

    def test_lenient_service(self):
        qualname_difference = _qualname(self.cls.difference)
        self.assertIn(qualname_difference, _LENIENT)
        self.assertTrue(_LENIENT[qualname_difference])
        self.assertTrue(_LENIENT[self.cls.difference])

    def test_cannot_differ_non_class(self):
        emsg = "Cannot differ"
        with self.assertRaisesRegex(TypeError, emsg):
            self.metadata.difference(None)

    def test_cannot_differ_different_class(self):
        other = CubeMetadata(*(None,) * len(CubeMetadata._fields))
        emsg = "Cannot differ"
        with self.assertRaisesRegex(TypeError, emsg):
            self.metadata.difference(other)

    def test_lenient_default(self):
        return_value = self.mock_kwargs.values()
        with mock.patch.object(
            self.cls, "_difference", return_value=return_value
        ) as mocker:
            result = self.metadata.difference(self.metadata)

        self.assertEqual(self.mock_kwargs, result._asdict())
        self.assertEqual(1, mocker.call_count)
        (arg,), kwargs = mocker.call_args
        self.assertEqual(id(self.metadata), id(arg))
        self.assertEqual(dict(), kwargs)

    def test_lenient_true(self):
        return_value = self.mock_kwargs.values()
        with mock.patch.object(
            self.cls, "_difference", return_value=return_value
        ) as mdifference:
            with mock.patch.object(_LENIENT, "context") as mcontext:
                result = self.metadata.difference(self.metadata, lenient=True)

        self.assertEqual(1, mcontext.call_count)
        (arg,), kwargs = mcontext.call_args
        self.assertEqual(_qualname(self.cls.difference), arg)
        self.assertEqual(dict(), kwargs)

        self.assertEqual(self.mock_kwargs, result._asdict())
        self.assertEqual(1, mdifference.call_count)
        (arg,), kwargs = mdifference.call_args
        self.assertEqual(id(self.metadata), id(arg))
        self.assertEqual(dict(), kwargs)

    def test_lenient_false(self):
        return_value = self.mock_kwargs.values()
        with mock.patch.object(
            self.cls, "_difference", return_value=return_value
        ) as mdifference:
            with mock.patch.object(_LENIENT, "context") as mcontext:
                result = self.metadata.difference(self.metadata, lenient=False)

        self.assertEqual(mcontext.call_count, 1)
        args, kwargs = mcontext.call_args
        self.assertEqual((), args)
        self.assertEqual({_qualname(self.cls.difference): False}, kwargs)

        self.assertEqual(self.mock_kwargs, result._asdict())
        self.assertEqual(1, mdifference.call_count)
        (arg,), kwargs = mdifference.call_args
        self.assertEqual(id(self.metadata), id(arg))
        self.assertEqual(dict(), kwargs)


class Test_equal(tests.IrisTest):
    def setUp(self):
        kwargs = dict(
            standard_name=sentinel.standard_name,
            long_name=sentinel.long_name,
            var_name=sentinel.var_name,
            units=sentinel.units,
            attributes=sentinel.attributes,
        )
        self.cls = BaseMetadata
        self.metadata = self.cls(**kwargs)

    def test_lenient_service(self):
        qualname_equal = _qualname(self.cls.equal)
        self.assertIn(qualname_equal, _LENIENT)
        self.assertTrue(_LENIENT[qualname_equal])
        self.assertTrue((_LENIENT[self.cls.equal]))

    def test_cannot_compare_non_class(self):
        emsg = "Cannot compare"
        with self.assertRaisesRegex(TypeError, emsg):
            self.metadata.equal(None)

    def test_cannot_compare_different_class(self):
        other = CubeMetadata(*(None,) * len(CubeMetadata._fields))
        emsg = "Cannot compare"
        with self.assertRaisesRegex(TypeError, emsg):
            self.metadata.equal(other)

    def test_lenient_default(self):
        return_value = sentinel.return_value
        with mock.patch.object(
            self.cls, "__eq__", return_value=return_value
        ) as mocker:
            result = self.metadata.equal(self.metadata)

        self.assertEqual(return_value, result)
        self.assertEqual(1, mocker.call_count)
        (arg,), kwargs = mocker.call_args
        self.assertEqual(id(self.metadata), id(arg))
        self.assertEqual(dict(), kwargs)

    def test_lenient_true(self):
        return_value = sentinel.return_value
        with mock.patch.object(
            self.cls, "__eq__", return_value=return_value
        ) as m__eq__:
            with mock.patch.object(_LENIENT, "context") as mcontext:
                result = self.metadata.equal(self.metadata, lenient=True)

        self.assertEqual(return_value, result)
        self.assertEqual(1, mcontext.call_count)
        (arg,), kwargs = mcontext.call_args
        self.assertEqual(_qualname(self.cls.equal), arg)
        self.assertEqual(dict(), kwargs)

        self.assertEqual(1, m__eq__.call_count)
        (arg,), kwargs = m__eq__.call_args
        self.assertEqual(id(self.metadata), id(arg))
        self.assertEqual(dict(), kwargs)

    def test_lenient_false(self):
        return_value = sentinel.return_value
        with mock.patch.object(
            self.cls, "__eq__", return_value=return_value
        ) as m__eq__:
            with mock.patch.object(_LENIENT, "context") as mcontext:
                result = self.metadata.equal(self.metadata, lenient=False)

        self.assertEqual(1, mcontext.call_count)
        args, kwargs = mcontext.call_args
        self.assertEqual((), args)
        self.assertEqual({_qualname(self.cls.equal): False}, kwargs)

        self.assertEqual(return_value, result)
        self.assertEqual(1, m__eq__.call_count)
        (arg,), kwargs = m__eq__.call_args
        self.assertEqual(id(self.metadata), id(arg))
        self.assertEqual(dict(), kwargs)


class Test_name(tests.IrisTest):
    def setUp(self):
        self.cls = BaseMetadata
        self.default = self.cls.DEFAULT_NAME

    @staticmethod
    def _make(standard_name=None, long_name=None, var_name=None):
        return BaseMetadata(
            standard_name=standard_name,
            long_name=long_name,
            var_name=var_name,
            units=None,
            attributes=None,
        )

    def test_standard_name(self):
        token = "standard_name"
        metadata = self._make(standard_name=token)

        result = metadata.name()
        self.assertEqual(token, result)
        result = metadata.name(token=True)
        self.assertEqual(token, result)

    def test_standard_name__invalid_token(self):
        token = "nope nope"
        metadata = self._make(standard_name=token)

        result = metadata.name()
        self.assertEqual(token, result)
        result = metadata.name(token=True)
        self.assertEqual(self.default, result)

    def test_long_name(self):
        token = "long_name"
        metadata = self._make(long_name=token)

        result = metadata.name()
        self.assertEqual(token, result)
        result = metadata.name(token=True)
        self.assertEqual(token, result)

    def test_long_name__invalid_token(self):
        token = "nope nope"
        metadata = self._make(long_name=token)

        result = metadata.name()
        self.assertEqual(token, result)
        result = metadata.name(token=True)
        self.assertEqual(self.default, result)

    def test_var_name(self):
        token = "var_name"
        metadata = self._make(var_name=token)

        result = metadata.name()
        self.assertEqual(token, result)
        result = metadata.name(token=True)
        self.assertEqual(token, result)

    def test_var_name__invalid_token(self):
        token = "nope nope"
        metadata = self._make(var_name=token)

        result = metadata.name()
        self.assertEqual(token, result)
        result = metadata.name(token=True)
        self.assertEqual(self.default, result)

    def test_default(self):
        metadata = self._make()

        result = metadata.name()
        self.assertEqual(self.default, result)
        result = metadata.name(token=True)
        self.assertEqual(self.default, result)

    def test_default__invalid_token(self):
        token = "nope nope"
        metadata = self._make()

        result = metadata.name(default=token)
        self.assertEqual(token, result)

        emsg = "Cannot retrieve a valid name token"
        with self.assertRaisesRegex(ValueError, emsg):
            metadata.name(default=token, token=True)


class Test_token(tests.IrisTest):
    def setUp(self):
        self.cls = BaseMetadata

    def test_passthru_None(self):
        result = self.cls.token(None)
        self.assertIsNone(result)

    def test_fail_leading_underscore(self):
        result = self.cls.token("_nope")
        self.assertIsNone(result)

    def test_fail_leading_dot(self):
        result = self.cls.token(".nope")
        self.assertIsNone(result)

    def test_fail_leading_plus(self):
        result = self.cls.token("+nope")
        self.assertIsNone(result)

    def test_fail_leading_at(self):
        result = self.cls.token("@nope")
        self.assertIsNone(result)

    def test_fail_space(self):
        result = self.cls.token("nope nope")
        self.assertIsNone(result)

    def test_fail_colon(self):
        result = self.cls.token("nope:")
        self.assertIsNone(result)

    def test_pass_simple(self):
        token = "simple"
        result = self.cls.token(token)
        self.assertEqual(token, result)

    def test_pass_leading_digit(self):
        token = "123simple"
        result = self.cls.token(token)
        self.assertEqual(token, result)

    def test_pass_mixture(self):
        token = "S.imple@one+two_3"
        result = self.cls.token(token)
        self.assertEqual(token, result)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""NAME file format loading functions."""

import collections
import datetime
import re
import warnings

import cf_units
import numpy as np

from iris.coords import AuxCoord, DimCoord, CellMethod
import iris.coord_systems
import iris.cube
from iris.exceptions import TranslationError
import iris.util

from operator import itemgetter


EARTH_RADIUS = 6371229.0
NAMEIII_DATETIME_FORMAT = "%d/%m/%Y  %H:%M %Z"
NAMETRAJ_DATETIME_FORMAT = "%d/%m/%Y  %H:%M:%S %Z"
NAMEII_FIELD_DATETIME_FORMAT = "%H%M%Z %d/%m/%Y"
NAMEII_TIMESERIES_DATETIME_FORMAT = "%d/%m/%Y  %H:%M:%S"


NAMECoord = collections.namedtuple(
    "NAMECoord", ["name", "dimension", "values"]
)


def _split_name_and_units(name):
    units = None
    if "(" in name and ")" in name:
        split = name.rsplit("(", 1)
        try_units = split[1].replace(")", "").strip()
        try:
            try_units = cf_units.Unit(try_units)
        except ValueError:
            pass
        else:
            name = split[0].strip()
            units = try_units
    return name, units


def read_header(file_handle):
    """
    Return a dictionary containing the header information extracted
    from the the provided NAME file object.

    Args:

    * file_handle (file-like object):
        A file-like object from which to read the header information.

    Returns:
        A dictionary containing the extracted header information.

    """
    header = {}
    header["NAME Version"] = next(file_handle).strip()
    for line in file_handle:
        words = line.split(":", 1)
        if len(words) != 2:
            if "Forward" in line or "Backward" in line:
                header["Trajectory direction"] = words[0].strip()
            break
        key, value = [word.strip() for word in words]
        header[key] = value

    # Cast some values into floats or integers if they match a
    # given name. Set any empty string values to None.
    for key, value in header.items():
        if value:
            if key in [
                "X grid origin",
                "Y grid origin",
                "X grid resolution",
                "Y grid resolution",
            ]:
                header[key] = float(value)
            elif key in [
                "X grid size",
                "Y grid size",
                "Number of preliminary cols",
                "Number of field cols",
                "Number of fields",
                "Number of series",
            ]:
                header[key] = int(value)
        else:
            header[key] = None

    return header


def _read_data_arrays(file_handle, n_arrays, shape):
    """
    Return a list of NumPy arrays containing the data extracted from
    the provided file object. The number and shape of the arrays
    must be specified.

    """
    data_arrays = [np.zeros(shape, dtype=np.float32) for i in range(n_arrays)]

    # Iterate over the remaining lines which represent the data in
    # a column form.
    for line in file_handle:
        # Split the line by comma, removing the last empty column
        # caused by the trailing comma
        vals = line.split(",")[:-1]

        # Cast the x and y grid positions to integers and convert
        # them to zero based indices
        x = int(float(vals[0])) - 1
        y = int(float(vals[1])) - 1

        # Populate the data arrays (i.e. all columns but the leading 4).
        for i, data_array in enumerate(data_arrays):
            data_array[y, x] = float(vals[i + 4])

    return data_arrays


def _build_lat_lon_for_NAME_field(
    header, dimindex, x_or_y, coord_names=["longitude", "latitude"]
):
    """
    Return regular latitude and longitude coordinates extracted from
    the provided header dictionary.
    """

    if x_or_y == "X":
        start = header["X grid origin"]
        step = header["X grid resolution"]
        count = header["X grid size"]
        pts = start + np.arange(count, dtype=np.float64) * step
        lat_lon = NAMECoord(
            name=coord_names[0], dimension=dimindex, values=pts
        )
    else:
        start = header["Y grid origin"]
        step = header["Y grid resolution"]
        count = header["Y grid size"]
        pts = start + np.arange(count, dtype=np.float64) * step
        lat_lon = NAMECoord(
            name=coord_names[1], dimension=dimindex, values=pts
        )

    return lat_lon


def _build_lat_lon_for_NAME_timeseries(column_headings):
    """
    Return regular latitude and longitude coordinates extracted from
    the provided column_headings dictionary.

    """
    pattern = re.compile(r"\-?[0-9]*\.[0-9]*")
    new_Xlocation_column_header = []
    for t in column_headings["X"]:
        if "Lat-Long" in t:
            matches = pattern.search(t)
            new_Xlocation_column_header.append(float(matches.group(0)))
        else:
            new_Xlocation_column_header.append(t)
    column_headings["X"] = new_Xlocation_column_header
    lon = NAMECoord(
        name="longitude", dimension=None, values=column_headings["X"]
    )

    new_Ylocation_column_header = []
    for t in column_headings["Y"]:
        if "Lat-Long" in t:
            matches = pattern.search(t)
            new_Ylocation_column_header.append(float(matches.group(0)))
        else:
            new_Ylocation_column_header.append(t)
    column_headings["Y"] = new_Ylocation_column_header
    lat = NAMECoord(
        name="latitude", dimension=None, values=column_headings["Y"]
    )

    return lat, lon


def _calc_integration_period(time_avgs):
    """
    Return a list of datetime.timedelta objects determined from the provided
    list of averaging/integration period column headings.

    """
    integration_periods = []
    pattern = re.compile(
        r"\s*(\d{1,2}day)?\s*(\d{1,2}hr)?\s*(\d{1,2}min)?\s*(\w*)\s*"
    )
    for time_str in time_avgs:
        days = 0
        hours = 0
        minutes = 0
        matches = pattern.search(time_str)
        if matches:
            _days = matches.group(1)
            if _days is not None and len(_days) > 0:
                days = float(_days.rstrip("day"))
            _hours = matches.group(2)
            if _hours is not None and len(_hours) > 0:
                hours = float(_hours.rstrip("hr"))
            _minutes = matches.group(3)
            if _minutes is not None and len(_minutes) > 0:
                minutes = float(_minutes.rstrip("min"))
        total_hours = days * 24.0 + hours + minutes / 60.0
        integration_periods.append(datetime.timedelta(hours=total_hours))
    return integration_periods


def _parse_units(units):
    """
    Return a known :class:`cf_units.Unit` given a NAME unit

    .. note::

        * Some NAME units are not currently handled.
        * Units which are in the wrong case (case is ignored in NAME)
        * Units where the space between SI units is missing
        * Units where the characters used are non-standard (i.e. 'mc' for
          micro instead of 'u')

    Args:

    * units (string):
        NAME units.

    Returns:
        An instance of :class:`cf_units.Unit`.

    """

    unit_mapper = {
        "Risks/m3": "1",  # Used for Bluetongue
        "TCID50s/m3": "1",  # Used for Foot and Mouth
        "TCID50/m3": "1",  # Used for Foot and Mouth
        "N/A": "1",  # Used for CHEMET area at risk
        "lb": "pounds",  # pounds
        "oz": "1",  # ounces
        "deg": "degree",  # angular degree
        "oktas": "1",  # oktas
        "deg C": "deg_C",  # degrees Celsius
        "FL": "unknown",  # flight level
    }

    units = unit_mapper.get(units, units)

    units = units.replace("Kg", "kg")
    units = units.replace("gs", "g s")
    units = units.replace("Bqs", "Bq s")
    units = units.replace("mcBq", "uBq")
    units = units.replace("mcg", "ug")
    try:
        units = cf_units.Unit(units)
    except ValueError:
        warnings.warn("Unknown units: {!r}".format(units))
        units = cf_units.Unit(None)

    return units


def _cf_height_from_name(z_coord, lower_bound=None, upper_bound=None):
    """
    Parser for the z component of field headings.

    This parse is specifically for handling the z component of NAME field
    headings, which include height above ground level, height above sea level
    and flight level etc.  This function returns an iris coordinate
    representing this field heading.

    Args:

    * z_coord (list):
        A field heading, specifically the z component.

    Returns:
        An instance of :class:`iris.coords.AuxCoord` representing the
        interpretation of the supplied field heading.

    """

    # NAMEII - integer/float support.
    # Match against height agl, asl and Pa.
    pattern = re.compile(
        r"^From\s*"
        r"(?P<lower_bound>[0-9]+(\.[0-9]+)?)"
        r"\s*-\s*"
        r"(?P<upper_bound>[0-9]+(\.[0-9]+)?)"
        r"\s*(?P<type>m\s*asl|m\s*agl|Pa)"
        r"(?P<extra>.*)"
    )

    # Match against flight level.
    pattern_fl = re.compile(
        r"^From\s*"
        r"(?P<type>FL)"
        r"(?P<lower_bound>[0-9]+(\.[0-9]+)?)"
        r"\s*-\s*FL"
        r"(?P<upper_bound>[0-9]+(\.[0-9]+)?)"
        r"(?P<extra>.*)"
    )

    # NAMEIII - integer/float support.
    # Match scalar against height agl, asl, Pa, FL
    pattern_scalar = re.compile(
        r"Z\s*=\s*"
        r"(?P<point>[0-9]+(\.[0-9]+)?([eE][+-]?\d+)?)"
        r"\s*(?P<type>m\s*agl|m\s*asl|FL|Pa)"
        r"(?P<extra>.*)"
    )

    type_name = {
        "magl": "height",
        "masl": "altitude",
        "FL": "flight_level",
        "Pa": "air_pressure",
    }
    patterns = [pattern, pattern_fl, pattern_scalar]

    units = "no-unit"
    points = z_coord
    bounds = None
    standard_name = None
    long_name = "z"

    if upper_bound is not None and lower_bound is not None:
        match_ub = pattern_scalar.match(upper_bound)
        match_lb = pattern_scalar.match(lower_bound)

    for pattern in patterns:
        match = pattern.match(z_coord)
        if match:
            match = match.groupdict()
            # Do not interpret if there is additional information to the match
            if match["extra"]:
                break
            units = match["type"].replace(" ", "")
            name = type_name[units]

            # Interpret points if present.
            if "point" in match:
                points = float(match["point"])
                if upper_bound is not None and lower_bound is not None:
                    bounds = np.array(
                        [
                            float(match_lb.groupdict()["point"]),
                            float(match_ub.groupdict()["point"]),
                        ]
                    )
            # Interpret points from bounds.
            else:
                bounds = np.array(
                    [float(match["lower_bound"]), float(match["upper_bound"])]
                )
                points = bounds.sum() / 2.0

            long_name = None
            if name == "altitude":
                units = units[0]
                standard_name = name
                long_name = "altitude above sea level"
            elif name == "height":
                units = units[0]
                standard_name = name
                long_name = "height above ground level"
            elif name == "air_pressure":
                standard_name = name
            elif name == "flight_level":
                long_name = name
            units = _parse_units(units)

            break

    coord = AuxCoord(
        points,
        units=units,
        standard_name=standard_name,
        long_name=long_name,
        bounds=bounds,
    )

    return coord


def _generate_cubes(
    header, column_headings, coords, data_arrays, cell_methods=None
):
    """
    Yield :class:`iris.cube.Cube` instances given
    the headers, column headings, coords and data_arrays extracted
    from a NAME file.

    """
    for i, data_array in enumerate(data_arrays):
        # Turn the dictionary of column headings with a list of header
        # information for each field into a dictionary of headings for
        # just this field.
        field_headings = {k: v[i] for k, v in column_headings.items()}

        # Make a cube.
        cube = iris.cube.Cube(data_array)

        # Determine the name and units.
        name = "{} {}".format(
            field_headings["Species"], field_headings["Quantity"]
        )
        name = name.upper().replace(" ", "_")
        cube.rename(name)

        # Some units are not in SI units, are missing spaces or typed
        # in the wrong case. _parse_units returns units that are
        # recognised by Iris.
        cube.units = _parse_units(field_headings["Units"])

        # Define and add the singular coordinates of the field (flight
        # level, time etc.)
        if "Z" in field_headings:
            (upper_bound,) = [
                field_headings["... to [Z]"]
                if "... to [Z]" in field_headings
                else None
            ]
            (lower_bound,) = [
                field_headings["... from [Z]"]
                if "... from [Z]" in field_headings
                else None
            ]
            z_coord = _cf_height_from_name(
                field_headings["Z"],
                upper_bound=upper_bound,
                lower_bound=lower_bound,
            )
            cube.add_aux_coord(z_coord)

        # Define the time unit and use it to serialise the datetime for
        # the time coordinate.
        time_unit = cf_units.Unit(
            "hours since epoch", calendar=cf_units.CALENDAR_GREGORIAN
        )

        # Build time, height, latitude and longitude coordinates.
        for coord in coords:
            pts = coord.values
            coord_sys = None
            if coord.name == "latitude" or coord.name == "longitude":
                coord_units = "degrees"
                coord_sys = iris.coord_systems.GeogCS(EARTH_RADIUS)
            if (
                coord.name == "projection_x_coordinate"
                or coord.name == "projection_y_coordinate"
            ):
                coord_units = "m"
                coord_sys = iris.coord_systems.OSGB()
            if coord.name == "height":
                coord_units = "m"
                long_name = "height above ground level"
                pts = coord.values
            if coord.name == "altitude":
                coord_units = "m"
                long_name = "altitude above sea level"
                pts = coord.values
            if coord.name == "air_pressure":
                coord_units = "Pa"
                pts = coord.values
            if coord.name == "flight_level":
                pts = coord.values
                long_name = "flight_level"
                coord_units = _parse_units("FL")
            if coord.name == "time":
                coord_units = time_unit
                pts = time_unit.date2num(coord.values)

            if coord.dimension is not None:
                if coord.name == "longitude":
                    circular = iris.util._is_circular(pts, 360.0)
                else:
                    circular = False
                if coord.name == "flight_level":
                    icoord = DimCoord(
                        points=pts, units=coord_units, long_name=long_name
                    )
                else:
                    icoord = DimCoord(
                        points=pts,
                        standard_name=coord.name,
                        units=coord_units,
                        coord_system=coord_sys,
                        circular=circular,
                    )
                if coord.name == "height" or coord.name == "altitude":
                    icoord.long_name = long_name
                if (
                    coord.name == "time"
                    and "Av or Int period" in field_headings
                ):
                    dt = coord.values - field_headings["Av or Int period"]
                    bnds = time_unit.date2num(np.vstack((dt, coord.values)).T)
                    icoord.bounds = bnds
                else:
                    icoord.guess_bounds()
                cube.add_dim_coord(icoord, coord.dimension)
            else:
                icoord = AuxCoord(
                    points=pts[i],
                    standard_name=coord.name,
                    coord_system=coord_sys,
                    units=coord_units,
                )
                if (
                    coord.name == "time"
                    and "Av or Int period" in field_headings
                ):
                    dt = coord.values - field_headings["Av or Int period"]
                    bnds = time_unit.date2num(np.vstack((dt, coord.values)).T)
                    icoord.bounds = bnds[i, :]
                cube.add_aux_coord(icoord)

        # Headings/column headings which are encoded elsewhere.
        headings = [
            "X",
            "Y",
            "Z",
            "Time",
            "T",
            "Units",
            "Av or Int period",
            "... from [Z]",
            "... to [Z]",
            "X grid origin",
            "Y grid origin",
            "X grid size",
            "Y grid size",
            "X grid resolution",
            "Y grid resolution",
            "Number of field cols",
            "Number of preliminary cols",
            "Number of fields",
            "Number of series",
            "Output format",
        ]

        # Add the Main Headings as attributes.
        for key, value in header.items():
            if value is not None and value != "" and key not in headings:
                cube.attributes[key] = value

        # Add the Column Headings as attributes
        for key, value in field_headings.items():
            if value is not None and value != "" and key not in headings:
                cube.attributes[key] = value

        if cell_methods is not None:
            cube.add_cell_method(cell_methods[i])

        yield cube


def _build_cell_methods(av_or_ints, coord):
    """
    Return a list of :class:`iris.coords.CellMethod` instances
    based on the provided list of column heading entries and the
    associated coordinate. If a given entry does not correspond to a cell
    method (e.g. "No time averaging"), a value of None is inserted.

    Args:

    * av_or_ints (iterable of strings):
        An iterable of strings containing the colummn heading entries
        to be parsed.
    * coord (string or :class:`iris.coords.Coord`):
        The coordinate name (or :class:`iris.coords.Coord` instance)
        to which the column heading entries refer.

    Returns:
        A list that is the same length as `av_or_ints` containing
        :class:`iris.coords.CellMethod` instances or values of None.

    """
    cell_methods = []
    no_avg_pattern = re.compile(r"^(no( (.* )?averaging)?)?$", re.IGNORECASE)
    for av_or_int in av_or_ints:
        if no_avg_pattern.search(av_or_int) is not None:
            cell_method = None
        elif "average" in av_or_int or "averaged" in av_or_int:
            cell_method = CellMethod("mean", coord)
        elif "integral" in av_or_int or "integrated" in av_or_int:
            cell_method = CellMethod("sum", coord)
        else:
            cell_method = None
            msg = "Unknown {} statistic: {!r}. Unable to create cell method."
            warnings.warn(msg.format(coord, av_or_int))
        cell_methods.append(cell_method)
    return cell_methods


def load_NAMEIII_field(filename):
    """
    Load a NAME III grid output file returning a
    generator of :class:`iris.cube.Cube` instances.

    Args:

    * filename (string):
        Name of file to load.

    Returns:
        A generator :class:`iris.cube.Cube` instances.

    """
    # Loading a file gives a generator of lines which can be progressed using
    # the next() function. This will come in handy as we wish to progress
    # through the file line by line.
    with open(filename, "r") as file_handle:
        # Create a dictionary which can hold the header metadata about this
        # file.
        header = read_header(file_handle)

        # Skip the next line (contains the word Fields:) in the file.
        next(file_handle)

        # Read the lines of column definitions.
        # In this version a fixed order of column headings is assumed (and
        # first 4 columns are ignored).
        column_headings = {}
        for column_header_name in [
            "Species Category",
            "Name",
            "Quantity",
            "Species",
            "Units",
            "Sources",
            "Ensemble Av",
            "Time Av or Int",
            "Horizontal Av or Int",
            "Vertical Av or Int",
            "Prob Perc",
            "Prob Perc Ens",
            "Prob Perc Time",
            "Time",
            "Z",
            "D",
        ]:
            cols = [col.strip() for col in next(file_handle).split(",")]
            column_headings[column_header_name] = cols[4:-1]

        # Read in the column titles to determine the coordinate system
        col_titles = next(file_handle).split(",")
        if "National Grid" in col_titles[2]:
            coord_names = [
                "projection_x_coordinate",
                "projection_y_coordinate",
            ]
        else:
            coord_names = ["longitude", "latitude"]

        # Convert the time to python datetimes.
        new_time_column_header = []
        for i, t in enumerate(column_headings["Time"]):
            dt = datetime.datetime.strptime(t, NAMEIII_DATETIME_FORMAT)
            new_time_column_header.append(dt)
        column_headings["Time"] = new_time_column_header

        # Convert averaging/integrating period to timedeltas.
        column_headings["Av or Int period"] = _calc_integration_period(
            column_headings["Time Av or Int"]
        )

        # Build a time coordinate.
        tdim = NAMECoord(
            name="time",
            dimension=None,
            values=np.array(column_headings["Time"]),
        )

        cell_methods = _build_cell_methods(
            column_headings["Time Av or Int"], tdim.name
        )

        # Build regular latitude and longitude coordinates.
        lon = _build_lat_lon_for_NAME_field(
            header, 1, "X", coord_names=coord_names
        )
        lat = _build_lat_lon_for_NAME_field(
            header, 0, "Y", coord_names=coord_names
        )

        coords = [lon, lat, tdim]

        # Create data arrays to hold the data for each column.
        n_arrays = header["Number of field cols"]
        shape = (header["Y grid size"], header["X grid size"])
        data_arrays = _read_data_arrays(file_handle, n_arrays, shape)

    return _generate_cubes(
        header, column_headings, coords, data_arrays, cell_methods
    )


def load_NAMEII_field(filename):
    """
    Load a NAME II grid output file returning a
    generator of :class:`iris.cube.Cube` instances.

    Args:

    * filename (string):
        Name of file to load.

    Returns:
        A generator :class:`iris.cube.Cube` instances.

    """
    with open(filename, "r") as file_handle:
        # Create a dictionary which can hold the header metadata about this
        # file.
        header = read_header(file_handle)

        # Origin in namever=2 format is bottom-left hand corner so alter this
        # to centre of a grid box
        header["X grid origin"] = (
            header["X grid origin"] + header["X grid resolution"] / 2
        )
        header["Y grid origin"] = (
            header["Y grid origin"] + header["Y grid resolution"] / 2
        )

        # Read the lines of column definitions.
        # In this version a fixed order of column headings is assumed (and
        # first 4 columns are ignored).
        column_headings = {}
        for column_header_name in [
            "Species Category",
            "Species",
            "Time Av or Int",
            "Quantity",
            "Units",
            "Z",
            "Time",
        ]:
            cols = [col.strip() for col in next(file_handle).split(",")]
            column_headings[column_header_name] = cols[4:-1]

        # Convert the time to python datetimes
        new_time_column_header = []
        for i, t in enumerate(column_headings["Time"]):
            dt = datetime.datetime.strptime(t, NAMEII_FIELD_DATETIME_FORMAT)
            new_time_column_header.append(dt)
        column_headings["Time"] = new_time_column_header

        # Convert averaging/integrating period to timedeltas.
        pattern = re.compile(r"\s*(\d{3})\s*(hr)?\s*(time)\s*(\w*)")
        column_headings["Av or Int period"] = []
        for i, t in enumerate(column_headings["Time Av or Int"]):
            matches = pattern.search(t)
            hours = 0
            if matches:
                if len(matches.group(1)) > 0:
                    hours = float(matches.group(1))
            column_headings["Av or Int period"].append(
                datetime.timedelta(hours=hours)
            )

        # Build a time coordinate.
        tdim = NAMECoord(
            name="time",
            dimension=None,
            values=np.array(column_headings["Time"]),
        )

        cell_methods = _build_cell_methods(
            column_headings["Time Av or Int"], tdim.name
        )

        # Build regular latitude and longitude coordinates.
        lon = _build_lat_lon_for_NAME_field(header, 1, "X")
        lat = _build_lat_lon_for_NAME_field(header, 0, "Y")

        coords = [lon, lat, tdim]

        # Skip the blank line after the column headings.
        next(file_handle)

        # Create data arrays to hold the data for each column.
        n_arrays = header["Number of fields"]
        shape = (header["Y grid size"], header["X grid size"])
        data_arrays = _read_data_arrays(file_handle, n_arrays, shape)

    return _generate_cubes(
        header, column_headings, coords, data_arrays, cell_methods
    )


def load_NAMEIII_timeseries(filename):
    """
    Load a NAME III time series file returning a
    generator of :class:`iris.cube.Cube` instances.

    Args:

    * filename (string):
        Name of file to load.

    Returns:
        A generator :class:`iris.cube.Cube` instances.

    """
    with open(filename, "r") as file_handle:
        # Create a dictionary which can hold the header metadata about this
        # file.
        header = read_header(file_handle)

        # skip the next line (contains the word Fields:) in the file.
        next(file_handle)

        # Read the lines of column definitions - currently hardwired
        column_headings = {}
        for column_header_name in [
            "Species Category",
            "Name",
            "Quantity",
            "Species",
            "Units",
            "Sources",
            "Ens Av",
            "Time Av or Int",
            "Horizontal Av or Int",
            "Vertical Av or Int",
            "Prob Perc",
            "Prob Perc Ens",
            "Prob Perc Time",
            "Location",
            "X",
            "Y",
            "Z",
            "D",
        ]:
            cols = [col.strip() for col in next(file_handle).split(",")]
            column_headings[column_header_name] = cols[1:-1]

        # Determine the coordinates of the data and store in namedtuples.
        # Extract latitude and longitude information from X, Y location
        # headings.
        lat, lon = _build_lat_lon_for_NAME_timeseries(column_headings)

        # Convert averaging/integrating period to timedeltas.
        column_headings["Av or Int period"] = _calc_integration_period(
            column_headings["Time Av or Int"]
        )

        # Skip the line after the column headings.
        next(file_handle)

        # Make a list of data lists to hold the data for each column.
        data_lists = [[] for i in range(header["Number of field cols"])]
        time_list = []

        # Iterate over the remaining lines which represent the data in a
        # column form.
        for line in file_handle:
            # Split the line by comma, removing the last empty column caused
            # by the trailing comma.
            vals = line.split(",")[:-1]

            # Time is stored in the first column.
            t = vals[0].strip()
            dt = datetime.datetime.strptime(t, NAMEIII_DATETIME_FORMAT)
            time_list.append(dt)

            # Populate the data arrays.
            for i, data_list in enumerate(data_lists):
                data_list.append(float(vals[i + 1]))

        data_arrays = [np.array(dl) for dl in data_lists]
        time_array = np.array(time_list)
        tdim = NAMECoord(name="time", dimension=0, values=time_array)

        coords = [lon, lat, tdim]

    return _generate_cubes(header, column_headings, coords, data_arrays)


def load_NAMEII_timeseries(filename):
    """
    Load a NAME II Time Series file returning a
    generator of :class:`iris.cube.Cube` instances.

    Args:

    * filename (string):
        Name of file to load.

    Returns:
        A generator :class:`iris.cube.Cube` instances.

    """
    with open(filename, "r") as file_handle:
        # Create a dictionary which can hold the header metadata about this
        # file.
        header = read_header(file_handle)

        # Read the lines of column definitions.
        column_headings = {}
        for column_header_name in [
            "Y",
            "X",
            "Location",
            "Species Category",
            "Species",
            "Quantity",
            "Z",
            "Units",
        ]:
            cols = [col.strip() for col in next(file_handle).split(",")]
            column_headings[column_header_name] = cols[1:-1]

        # Determine the coordinates of the data and store in namedtuples.
        # Extract latitude and longitude information from X, Y location
        # headings.
        lat, lon = _build_lat_lon_for_NAME_timeseries(column_headings)

        # Skip the blank line after the column headings.
        next(file_handle)

        # Make a list of data arrays to hold the data for each column.
        data_lists = [[] for i in range(header["Number of series"])]
        time_list = []

        # Iterate over the remaining lines which represent the data in a
        # column form.
        for line in file_handle:
            # Split the line by comma, removing the last empty column caused
            # by the trailing comma.
            vals = line.split(",")[:-1]

            # Time is stored in the first two columns.
            t = vals[0].strip() + " " + vals[1].strip()
            dt = datetime.datetime.strptime(
                t, NAMEII_TIMESERIES_DATETIME_FORMAT
            )
            time_list.append(dt)

            # Populate the data arrays.
            for i, data_list in enumerate(data_lists):
                data_list.append(float(vals[i + 2]))

        data_arrays = [np.array(dl) for dl in data_lists]
        time_array = np.array(time_list)
        tdim = NAMECoord(name="time", dimension=0, values=time_array)

        coords = [lon, lat, tdim]

    return _generate_cubes(header, column_headings, coords, data_arrays)


def load_NAMEIII_version2(filename):
    """
    Load a NAME III version 2 file returning a
    generator of :class:`iris.cube.Cube` instances.

    Args:

    * filename (string):
        Name of file to load.

    Returns:
        A generator :class:`iris.cube.Cube` instances.

    """

    # loading a file gives a generator of lines which can be progressed
    # using the next() method. This will come in handy as we wish to
    # progress through the file line by line.
    with open(filename, "r") as file_handle:

        # define a dictionary to hold the header metadata about this file
        header = read_header(file_handle)

        # Skip next line which contains (Fields:)
        next(file_handle)

        # Now carry on and read column headers
        column_headings = {}
        datacol1 = header["Number of preliminary cols"]
        for line in file_handle:

            data = [col.strip() for col in line.split(",")][:-1]

            # If first column is not zero we have reached the end
            #  of the headers
            if data[0] != "":
                break

            column_key = data[datacol1 - 1].strip(":")

            # This will filter out any zero columns
            if filter(None, data[datacol1:]):
                column_headings[column_key] = data[datacol1:]

        # Some tidying up
        if "T" in column_headings:
            new_time_column_header = []
            for i, t in enumerate(column_headings["T"]):
                dt = datetime.datetime.strptime(t, NAMEIII_DATETIME_FORMAT)
                new_time_column_header.append(dt)
            column_headings["T"] = new_time_column_header

        # Convert averaging/integrating period to timedeltas.
        column_headings["Av or Int period"] = _calc_integration_period(
            column_headings["Time av/int info"]
        )

        # Next we need to figure out what we have in the preliminary columns
        # For X and Y we want the index
        # And the values can be extracted from the header information
        xindex = None
        yindex = None
        dim_coords = []

        # First determine whether we are using National Grid or lat/lon
        if "X (UK National Grid (m))" in data:
            coord_names = [
                "projection_x_coordinate",
                "projection_y_coordinate",
            ]
        else:
            coord_names = ["longitude", "latitude"]

        if "Y Index" in data:
            yindex = data.index("Y Index")
            dim_coords.append("Y")
            lat = _build_lat_lon_for_NAME_field(
                header, dim_coords.index("Y"), "Y", coord_names=coord_names
            )
        if "X Index" in data:
            xindex = data.index("X Index")
            dim_coords.append("X")
            lon = _build_lat_lon_for_NAME_field(
                header, dim_coords.index("X"), "X", coord_names=coord_names
            )

        # For all other variables we need the values (note that for Z the units
        # will also be given in the column header)
        tindex = None
        zindex = None
        if "T" in data:
            tindex = data.index("T")
            dim_coords.append("T")

        if "Z" in line:
            zgrid = [item for item in data if item[0:3] == "Z ("]
            zunits = zgrid[0].split("(")[1].strip(")")
            if zunits == "m asl":
                z_name = "altitude"
            elif zunits == "m agl":
                z_name = "height"
            elif zunits == "FL":
                z_name = "flight_level"
            elif zunits == "Pa":
                z_name = "air_pressure"
            else:
                ValueError("Vertical coordinate unkown")
            zindex = data.index(zgrid[0])
            dim_coords.append("Z")

        # Make a list of data lists to hold the data
        # for each column.(aimed at T-Z data)
        data_lists = [[] for i in range(header["Number of field cols"])]
        coord_lists = [
            [] for i in range(header["Number of preliminary cols"] - 1)
        ]

        # Iterate over the remaining lines which represent the data in a
        # column form.
        for line in file_handle:
            # Split the line by comma, removing the last empty column caused
            # by the trailing comma.
            vals = line.split(",")[:-1]

            # Time is stored in the column labelled T index
            if tindex is not None:
                t = vals[tindex].strip()
                dt = datetime.datetime.strptime(t, NAMEIII_DATETIME_FORMAT)
                coord_lists[dim_coords.index("T")].append(dt)

            # Z is stored in the column labelled ZIndex
            if zindex is not None:
                z = vals[zindex].strip()
                coord_lists[dim_coords.index("Z")].append(float(z))

            # For X and Y we are extracting indices not values
            if yindex is not None:
                yind = vals[yindex].strip()
                coord_lists[dim_coords.index("Y")].append(int(yind) - 1)
            if xindex is not None:
                xind = vals[xindex].strip()
                coord_lists[dim_coords.index("X")].append(int(xind) - 1)

            # Populate the data arrays.
            for i, data_list in enumerate(data_lists):
                data_list.append(float(vals[i + datacol1]))

        data_arrays = [np.array(dl) for dl in data_lists]

        # Convert Z and T arrays into arrays of indices
        zind = []
        if zindex is not None:
            z_array = np.array(coord_lists[dim_coords.index("Z")])
            z_unique = sorted(list(set(coord_lists[dim_coords.index("Z")])))
            z_coord = NAMECoord(
                name=z_name, dimension=dim_coords.index("Z"), values=z_unique
            )
            for z in z_array:
                zind.append(z_unique.index(z))
            coord_lists[dim_coords.index("Z")] = zind

        tind = []
        if tindex is not None:
            time_array = np.array(coord_lists[dim_coords.index("T")])
            t_unique = sorted(list(set(coord_lists[dim_coords.index("T")])))
            time = NAMECoord(
                name="time",
                dimension=dim_coords.index("T"),
                values=np.array(t_unique),
            )
            for t in time_array:
                tind.append(t_unique.index(t))
            coord_lists[dim_coords.index("T")] = tind

        # Now determine the shape of the multidimensional array to store
        # the data in based on the length of the coordinates
        array_shape_list = []
        coords = []
        for cname in dim_coords:
            if cname == "X":
                coords.append(lon)
                array_shape_list.append(len(lon.values))
            elif cname == "Y":
                coords.append(lat)
                array_shape_list.append(len(lat.values))
            elif cname == "Z":
                coords.append(z_coord)
                array_shape_list.append(len(z_coord.values))
            elif cname == "T":
                coords.append(time)
                array_shape_list.append(len(time.values))
        array_shape = np.array(array_shape_list)

        # Reshape the data to the new multidimensional shape
        new_data_arrays = []
        for data_array in data_arrays:
            new_data_array = np.zeros(array_shape, dtype=np.float32)
            for ind1, item in enumerate(data_array):
                index_list = []
                for column, dcoord in enumerate(dim_coords):
                    index_list.append(coord_lists[column][ind1])
                index_array = np.array(index_list)
                mindex = np.ravel_multi_index(index_array, array_shape)
                new_data_array[np.unravel_index(mindex, array_shape)] = item
            new_data_arrays.append(new_data_array)

    # If X and Y are in the column headings build coordinates
    if "X" in column_headings and "Y" in column_headings:
        lat, lon = _build_lat_lon_for_NAME_timeseries(column_headings)
        coords.append(lat)
        coords.append(lon)

    # If time is in the column heading build a coordinate
    if "T" in column_headings:
        tdim = NAMECoord(
            name="time", dimension=None, values=np.array(column_headings["T"])
        )
        coords.append(tdim)

    return _generate_cubes(
        header, column_headings, coords, new_data_arrays, cell_methods=None
    )


def load_NAMEIII_trajectory(filename):
    """
    Load a NAME III trajectory file returning a
    generator of :class:`iris.cube.Cube` instances.

    Args:

    * filename (string):
        Name of file to load.

    Returns:
        A generator :class:`iris.cube.Cube` instances.

    """
    time_unit = cf_units.Unit(
        "hours since epoch", calendar=cf_units.CALENDAR_GREGORIAN
    )

    with open(filename, "r") as infile:
        header = read_header(infile)

        # read the column headings
        for line in infile:
            if line.startswith("    "):
                break
        headings = [heading.strip() for heading in line.split(",")]

        # read the columns
        columns = [[] for i in range(len(headings))]
        for line in infile:
            values = [v.strip() for v in line.split(",")]
            for c, v in enumerate(values):
                if "UTC" in v:
                    v = datetime.datetime.strptime(v, NAMETRAJ_DATETIME_FORMAT)
                else:
                    try:
                        v = float(v)
                    except ValueError:
                        pass
                columns[c].append(v)

    # Sort columns according to PP Index
    columns_t = list(map(list, zip(*columns)))
    columns_t.sort(key=itemgetter(1))
    columns = list(map(list, zip(*columns_t)))

    # Where's the Z column?
    z_column = None
    for i, heading in enumerate(headings):
        if heading.startswith("Z "):
            z_column = i
            break
    if z_column is None:
        raise TranslationError("Expected a Z column")

    # Every column up to Z becomes a coordinate.
    coords = []
    for name, values in zip(headings[: z_column + 1], columns[: z_column + 1]):
        values = np.array(values)
        if np.all(np.array(values) == values[0]):
            values = [values[0]]

        long_name = units = None
        if isinstance(values[0], datetime.datetime):
            values = time_unit.date2num(values)
            units = time_unit
            if name == "Time":
                name = "time"
        elif " (Lat-Long)" in name:
            if name.startswith("X"):
                name = "longitude"
            elif name.startswith("Y"):
                name = "latitude"
            units = "degrees"
        elif name == "Z (m asl)":
            name = "altitude"
            units = "m"
            long_name = "altitude above sea level"
        elif name == "Z (m agl)":
            name = "height"
            units = "m"
            long_name = "height above ground level"
        elif name == "Z (FL)":
            name = "flight_level"
            long_name = name

        try:
            coord = DimCoord(values, units=units)
        except ValueError:
            coord = AuxCoord(values, units=units)
        coord.rename(name)
        if coord.long_name is None and long_name is not None:
            coord.long_name = long_name
        coords.append(coord)

    # Every numerical column after the Z becomes a cube.
    for name, values in zip(headings[z_column + 1 :], columns[z_column + 1 :]):
        try:
            float(values[0])
        except ValueError:
            continue
        # units embedded in column heading?
        name, units = _split_name_and_units(name)
        cube = iris.cube.Cube(values, units=units)
        cube.rename(name)
        # Add the Main Headings as attributes.
        for key, value in header.items():
            if value is not None and value != "" and key not in headings:
                cube.attributes[key] = value
        # Add coordinates
        for coord in coords:
            dim = 0 if len(coord.points) > 1 else None
            if dim == 0 and coord.name() == "time":
                cube.add_dim_coord(coord.copy(), dim)
            elif dim == 0 and coord.name() == "PP Index":
                cube.add_dim_coord(coord.copy(), dim)
            else:
                cube.add_aux_coord(coord.copy(), dim)
        yield cube



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.util.squeeze`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import unittest

import iris
import iris.tests.stock as stock


class Test(tests.IrisTest):
    def setUp(self):
        self.cube = stock.simple_2d_w_multidim_and_scalars()

    def test_no_change(self):
        self.assertEqual(self.cube, iris.util.squeeze(self.cube))

    def test_squeeze_one_dim(self):
        cube_3d = iris.util.new_axis(self.cube, scalar_coord="an_other")
        cube_2d = iris.util.squeeze(cube_3d)

        self.assertEqual(self.cube, cube_2d)

    def test_squeeze_two_dims(self):
        cube_3d = iris.util.new_axis(self.cube, scalar_coord="an_other")
        cube_4d = iris.util.new_axis(cube_3d, scalar_coord="air_temperature")

        self.assertEqual(self.cube, iris.util.squeeze(cube_4d))

    def test_squeeze_one_anonymous_dim(self):
        cube_3d = iris.util.new_axis(self.cube)
        cube_2d = iris.util.squeeze(cube_3d)

        self.assertEqual(self.cube, cube_2d)

    def test_squeeze_to_scalar_cube(self):
        cube_scalar = self.cube[0, 0]
        cube_1d = iris.util.new_axis(cube_scalar)

        self.assertEqual(cube_scalar, iris.util.squeeze(cube_1d))


if __name__ == "__main__":
    unittest.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris._lazy data.non_lazy`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris._lazy_data import as_lazy_data, is_lazy_data, non_lazy


class Test_non_lazy(tests.IrisTest):
    def setUp(self):
        self.array = np.arange(8).reshape(2, 4)
        self.lazy_array = as_lazy_data(self.array)
        self.func = non_lazy(lambda array: array.sum(axis=0))
        self.func_result = [4, 6, 8, 10]

    def test_lazy_input(self):
        result = self.func(self.lazy_array)
        self.assertFalse(is_lazy_data(result))
        self.assertArrayEqual(result, self.func_result)

    def test_non_lazy_input(self):
        # Check that a non-lazy input doesn't trip up the functionality.
        result = self.func(self.array)
        self.assertFalse(is_lazy_data(result))
        self.assertArrayEqual(result, self.func_result)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :func:`iris.util.equalise_attributes` function.

"""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

import numpy as np

from iris.cube import Cube
from iris.util import equalise_attributes
import iris.tests.stock


class TestEqualiseAttributes(tests.IrisTest):
    def setUp(self):
        empty = Cube([])

        self.cube_no_attrs = empty.copy()

        self.cube_a1 = empty.copy()
        self.cube_a1.attributes.update({"a": 1})

        self.cube_a2 = empty.copy()
        self.cube_a2.attributes.update({"a": 2})

        self.cube_a1b5 = empty.copy()
        self.cube_a1b5.attributes.update({"a": 1, "b": 5})

        self.cube_a1b6 = empty.copy()
        self.cube_a1b6.attributes.update({"a": 1, "b": 6})

        self.cube_a2b6 = empty.copy()
        self.cube_a2b6.attributes.update({"a": 2, "b": 6})

        self.cube_b5 = empty.copy()
        self.cube_b5.attributes.update({"b": 5})

        # Array attribute values
        v1 = np.array([11, 12, 13])
        v2 = np.array([11, 9999, 13])
        self.v1 = v1

        self.cube_a1b5v1 = empty.copy()
        self.cube_a1b5v1.attributes.update({"a": 1, "b": 5, "v": v1})

        self.cube_a1b6v1 = empty.copy()
        self.cube_a1b6v1.attributes.update({"a": 1, "b": 6, "v": v1})

        self.cube_a1b6v2 = empty.copy()
        self.cube_a1b6v2.attributes.update({"a": 1, "b": 6, "v": v2})

    def _test(self, cubes, expect_attributes):
        """Test."""
        working_cubes = [cube.copy() for cube in cubes]
        original_working_list = [cube for cube in working_cubes]
        # Exercise basic operation
        equalise_attributes(working_cubes)
        # Check they are the same cubes
        self.assertEqual(working_cubes, original_working_list)
        # Check resulting attributes all match the expected set
        for cube in working_cubes:
            self.assertEqual(cube.attributes, expect_attributes)
        # Check everything else remains the same
        for new_cube, old_cube in zip(working_cubes, cubes):
            cube_before_noatts = old_cube.copy()
            cube_before_noatts.attributes.clear()
            cube_after_noatts = new_cube.copy()
            cube_after_noatts.attributes.clear()
            self.assertEqual(cube_after_noatts, cube_before_noatts)

    def test_no_attrs(self):
        cubes = [self.cube_no_attrs]
        self._test(cubes, {})

    def test_single(self):
        cubes = [self.cube_a1]
        self._test(cubes, {"a": 1})

    def test_identical(self):
        cubes = [self.cube_a1, self.cube_a1.copy()]
        self._test(cubes, {"a": 1})

    def test_one_extra(self):
        cubes = [self.cube_a1, self.cube_a1b5.copy()]
        self._test(cubes, {"a": 1})

    def test_one_different(self):
        cubes = [self.cube_a1b5, self.cube_a1b6]
        self._test(cubes, {"a": 1})

    def test_common_no_diffs(self):
        cubes = [self.cube_a1b5, self.cube_a1b5.copy()]
        self._test(cubes, {"a": 1, "b": 5})

    def test_common_all_diffs(self):
        cubes = [self.cube_a1b5, self.cube_a2b6]
        self._test(cubes, {})

    def test_none_common(self):
        cubes = [self.cube_a1, self.cube_b5]
        self._test(cubes, {})

    def test_array_extra(self):
        cubes = [self.cube_a1b6, self.cube_a1b6v1]
        self._test(cubes, {"a": 1, "b": 6})

    def test_array_different(self):
        cubes = [self.cube_a1b5v1, self.cube_a1b6v2]
        self._test(cubes, {"a": 1})

    def test_array_same(self):
        cubes = [self.cube_a1b5v1, self.cube_a1b6v1]
        self._test(cubes, {"a": 1, "v": self.v1})

    @tests.skip_data
    def test_complex_nonecommon(self):
        # Example with cell methods and factories, but no common attributes.
        cubes = [
            iris.tests.stock.global_pp(),
            iris.tests.stock.hybrid_height(),
        ]
        self._test(cubes, {})

    @tests.skip_data
    def test_complex_somecommon(self):
        # Example with cell methods and factories, plus some common attributes.
        cubes = [iris.tests.stock.global_pp(), iris.tests.stock.simple_pp()]
        self._test(
            cubes,
            {
                "STASH": iris.fileformats.pp.STASH(
                    model=1, section=16, item=203
                ),
                "source": "Data from Met Office Unified Model",
            },
        )


if __name__ == "__main__":
    tests.main()



"""
Loading a Cube From a Custom File Format
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This example shows how a custom text file can be loaded using the standard Iris
load mechanism.

The first stage in the process is to define an Iris :class:`FormatSpecification
<iris.io.format_picker.FormatSpecification>` for the file format. To create a
format specification we need to define the following:

* format_name - Some text that describes the format specification we are
  creating
* file_element - FileElement object describing the element which identifies
  this FormatSpecification.

  Possible values are:

    ``iris.io.format_picker.MagicNumber(n, o)``
        The n bytes from the file at offset o.

    ``iris.io.format_picker.FileExtension()``
        The file's extension.

    ``iris.io.format_picker.LeadingLine()``
        The first line of the file.

* file_element_value - The value that the file_element should take if a file
  matches this FormatSpecification
* handler (optional) - A generator function that will be called when the file
  specification has been identified. This function is provided by the user and
  provides the means to parse the whole file. If no handler function is
  provided, then identification is still possible without any handling.

  The handler function must define the following arguments:

  * list of filenames to process
  * callback function - An optional function to filter/alter the Iris cubes
    returned

  The handler function must be defined as generator which yields each cube as
  they are produced.

* priority (optional) - Integer giving a priority for considering this
  specification where higher priority means sooner consideration

In the following example, the function :func:`load_NAME_III` has been defined
to handle the loading of the raw data from the custom file format. This
function is called from :func:`NAME_to_cube` which uses this data to create and
yield Iris cubes.

In the ``main()`` function the filenames are loaded via the ``iris.load_cube``
function which automatically invokes the ``FormatSpecification`` we defined.
The cube returned from the load function is then used to produce a plot.

"""

import datetime

from cf_units import Unit, CALENDAR_GREGORIAN
import matplotlib.pyplot as plt
import numpy as np

import iris
import iris.coords as icoords
import iris.coord_systems as icoord_systems
import iris.fileformats
import iris.io.format_picker as format_picker
import iris.plot as iplt


UTC_format = "%H%M%Z %d/%m/%Y"

FLOAT_HEADERS = [
    "X grid origin",
    "Y grid origin",
    "X grid resolution",
    "Y grid resolution",
]
INT_HEADERS = ["X grid size", "Y grid size", "Number of fields"]
DATE_HEADERS = ["Run time", "Start of release", "End of release"]
COLUMN_NAMES = [
    "species_category",
    "species",
    "cell_measure",
    "quantity",
    "unit",
    "z_level",
    "time",
]


def load_NAME_III(filename):
    """
    Loads the Met Office's NAME III grid output files returning headers, column
    definitions and data arrays as 3 separate lists.

    """

    # Loading a file gives a generator of lines which can be progressed using
    # the next() function. This will come in handy as we wish to progress
    # through the file line by line.
    with open(filename) as file_handle:
        # Define a dictionary which can hold the header metadata for this file.
        headers = {}

        # Skip the NAME header of the file which looks something like
        # 'NAME III (version X.X.X)'.
        next(file_handle)

        # Read the next 16 lines of header information, putting the form
        # "header name:    header value" into a dictionary.
        for _ in range(16):
            header_name, header_value = next(file_handle).split(":")

            # Strip off any spurious space characters in the header name and
            # value.
            header_name = header_name.strip()
            header_value = header_value.strip()

            # Cast some headers into floats or integers if they match a given
            # header name.
            if header_name in FLOAT_HEADERS:
                header_value = float(header_value)
            elif header_name in INT_HEADERS:
                header_value = int(header_value)
            elif header_name in DATE_HEADERS:
                # convert the time to python datetimes
                header_value = datetime.datetime.strptime(
                    header_value, UTC_format
                )

            headers[header_name] = header_value

        # Skip the next blank line in the file.
        next(file_handle)

        # Read the next 7 lines of column definitions.
        column_headings = {}
        for column_header_name in COLUMN_NAMES:
            column_headings[column_header_name] = [
                col.strip() for col in next(file_handle).split(",")
            ][:-1]

        # Convert the time to python datetimes.
        new_time_column_header = []
        for i, t in enumerate(column_headings["time"]):
            # The first 4 columns aren't time at all, so don't convert them to
            # datetimes.
            if i >= 4:
                t = datetime.datetime.strptime(t, UTC_format)
            new_time_column_header.append(t)
        column_headings["time"] = new_time_column_header

        # Skip the blank line after the column headers.
        next(file_handle)

        # Make a list of data arrays to hold the data for each column.
        data_shape = (headers["Y grid size"], headers["X grid size"])
        data_arrays = [
            np.zeros(data_shape, dtype=np.float32)
            for i in range(headers["Number of fields"])
        ]

        # Iterate over the remaining lines which represent the data in a column
        # form.
        for line in file_handle:
            # Split the line by comma, removing the last empty column caused by
            # the trailing comma.
            vals = line.split(",")[:-1]

            # Cast the x and y grid positions to floats and convert them to
            # zero based indices (the numbers are 1 based grid positions where
            # 0.5 represents half a grid point.)
            x = int(float(vals[0]) - 1.5)
            y = int(float(vals[1]) - 1.5)

            # Populate the data arrays (i.e. all columns but the leading 4).
            for i, data_array in enumerate(data_arrays):
                data_array[y, x] = float(vals[i + 4])

    return headers, column_headings, data_arrays


def NAME_to_cube(filenames, callback):
    """
    Returns a generator of cubes given a list of filenames and a callback.
    """

    for filename in filenames:
        header, column_headings, data_arrays = load_NAME_III(filename)

        for i, data_array in enumerate(data_arrays):
            # turn the dictionary of column headers with a list of header
            # information for each field into a dictionary of headers for just
            # this field. Ignore the first 4 columns of grid position (data was
            # located with the data array).
            field_headings = dict(
                (k, v[i + 4]) for k, v in column_headings.items()
            )

            # make an cube
            cube = iris.cube.Cube(data_array)

            # define the name and unit
            name = "%s %s" % (
                field_headings["species"],
                field_headings["quantity"],
            )
            name = name.upper().replace(" ", "_")
            cube.rename(name)
            # Some units are badly encoded in the file, fix this by putting a
            # space in between. (if gs is not found, then the string will be
            # returned unchanged)
            cube.units = field_headings["unit"].replace("gs", "g s")

            # define and add the singular coordinates of the field (flight
            # level, time etc.)
            cube.add_aux_coord(
                icoords.AuxCoord(
                    field_headings["z_level"],
                    long_name="flight_level",
                    units="1",
                )
            )

            # define the time unit and use it to serialise the datetime for the
            # time coordinate
            time_unit = Unit("hours since epoch", calendar=CALENDAR_GREGORIAN)
            time_coord = icoords.AuxCoord(
                time_unit.date2num(field_headings["time"]),
                standard_name="time",
                units=time_unit,
            )
            cube.add_aux_coord(time_coord)

            # build a coordinate system which can be referenced by latitude and
            # longitude coordinates
            lat_lon_coord_system = icoord_systems.GeogCS(6371229)

            # build regular latitude and longitude coordinates which have
            # bounds
            start = header["X grid origin"] + header["X grid resolution"]
            step = header["X grid resolution"]
            count = header["X grid size"]
            pts = start + np.arange(count, dtype=np.float32) * step
            lon_coord = icoords.DimCoord(
                pts,
                standard_name="longitude",
                units="degrees",
                coord_system=lat_lon_coord_system,
            )
            lon_coord.guess_bounds()

            start = header["Y grid origin"] + header["Y grid resolution"]
            step = header["Y grid resolution"]
            count = header["Y grid size"]
            pts = start + np.arange(count, dtype=np.float32) * step
            lat_coord = icoords.DimCoord(
                pts,
                standard_name="latitude",
                units="degrees",
                coord_system=lat_lon_coord_system,
            )
            lat_coord.guess_bounds()

            # add the latitude and longitude coordinates to the cube, with
            # mappings to data dimensions
            cube.add_dim_coord(lat_coord, 0)
            cube.add_dim_coord(lon_coord, 1)

            # implement standard iris callback capability. Although callbacks
            # are not used in this example, the standard mechanism for a custom
            # loader to implement a callback is shown:
            cube = iris.io.run_callback(
                callback, cube, [header, field_headings, data_array], filename
            )

            # yield the cube created (the loop will continue when the next()
            # element is requested)
            yield cube


# Create a format_picker specification of the NAME file format giving it a
# priority greater than the built in NAME loader.
_NAME_III_spec = format_picker.FormatSpecification(
    "Name III",
    format_picker.LeadingLine(),
    lambda line: line.startswith(b"NAME III"),
    NAME_to_cube,
    priority=6,
)

# Register the NAME loader with iris
iris.fileformats.FORMAT_AGENT.add_spec(_NAME_III_spec)


# ---------------------------------------------
# |          Using the new loader             |
# ---------------------------------------------


def main():
    fname = iris.sample_data_path("NAME_output.txt")

    boundary_volc_ash_constraint = iris.Constraint(
        "VOLCANIC_ASH_AIR_CONCENTRATION", flight_level="From FL000 - FL200"
    )

    # Callback shown as None to illustrate where a cube-level callback function
    # would be used if required
    cube = iris.load_cube(fname, boundary_volc_ash_constraint, callback=None)

    # draw contour levels for the data (the top level is just a catch-all)
    levels = (0.0002, 0.002, 0.004, 1e10)
    cs = iplt.contourf(
        cube,
        levels=levels,
        colors=("#80ffff", "#939598", "#e00404"),
    )

    # draw a black outline at the lowest contour to highlight affected areas
    iplt.contour(cube, levels=(levels[0], 100), colors="black")

    # set an extent and a background image for the map
    ax = plt.gca()
    ax.set_extent((-90, 20, 20, 75))
    ax.stock_img("ne_shaded")

    # make a legend, with custom labels, for the coloured contour set
    artists, _ = cs.legend_elements()
    labels = [
        r"$%s < x \leq %s$" % (levels[0], levels[1]),
        r"$%s < x \leq %s$" % (levels[1], levels[2]),
        r"$x > %s$" % levels[2],
    ]
    ax.legend(
        artists, labels, title="Ash concentration / g m-3", loc="upper left"
    )

    time = cube.coord("time")
    time_date = time.units.num2date(time.points[0]).strftime(UTC_format)
    plt.title("Volcanic ash concentration forecast\nvalid at %s" % time_date)

    iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

from copy import deepcopy
import os
from types import GeneratorType
import unittest
from unittest import mock

import cftime
from numpy.testing import assert_array_equal

import iris.fileformats
import iris.fileformats.pp as pp
import iris.util


@tests.skip_data
class TestPPCopy(tests.IrisTest):
    def setUp(self):
        self.filename = tests.get_data_path(("PP", "aPPglob1", "global.pp"))

    def test_copy_field_deferred(self):
        field = next(pp.load(self.filename))
        clone = field.copy()
        self.assertEqual(field, clone)
        clone.lbyr = 666
        self.assertNotEqual(field, clone)

    def test_deepcopy_field_deferred(self):
        field = next(pp.load(self.filename))
        clone = deepcopy(field)
        self.assertEqual(field, clone)
        clone.lbyr = 666
        self.assertNotEqual(field, clone)

    def test_copy_field_non_deferred(self):
        field = next(pp.load(self.filename, True))
        clone = field.copy()
        self.assertEqual(field, clone)
        clone.data[0][0] = 666
        self.assertNotEqual(field, clone)

    def test_deepcopy_field_non_deferred(self):
        field = next(pp.load(self.filename, True))
        clone = deepcopy(field)
        self.assertEqual(field, clone)
        clone.data[0][0] = 666
        self.assertNotEqual(field, clone)


class IrisPPTest(tests.IrisTest):
    def check_pp(self, pp_fields, reference_filename):
        """
        Checks the given iterable of PPField objects matches the reference file, or creates the
        reference file if it doesn't exist.

        """
        # turn the generator into a list
        pp_fields = list(pp_fields)

        # Load deferred data for all of the fields (but don't do anything with it)
        for pp_field in pp_fields:
            pp_field.data

        test_string = str(pp_fields)
        reference_path = tests.get_result_path(reference_filename)
        if os.path.isfile(reference_path):
            with open(reference_path, "r") as reference_fh:
                reference = "".join(reference_fh.readlines())
            self._assert_str_same(
                reference + "\n",
                test_string + "\n",
                reference_filename,
                type_comparison_name="PP files",
            )
        else:
            with open(reference_path, "w") as reference_fh:
                reference_fh.writelines(test_string)


class TestPPHeaderDerived(tests.IrisTest):
    def setUp(self):
        self.pp = pp.PPField2()
        self.pp.lbuser = (0, 1, 2, 3, 4, 5, 6)
        self.pp.lbtim = 11
        self.pp.lbproc = 65539

    def test_standard_access(self):
        self.assertEqual(self.pp.lbtim, 11)

    def test_lbtim_access(self):
        self.assertEqual(self.pp.lbtim[0], 1)
        self.assertEqual(self.pp.lbtim.ic, 1)

    def test_lbtim_setter(self):
        self.pp.lbtim[4] = 4
        self.pp.lbtim[0] = 4
        self.assertEqual(self.pp.lbtim[0], 4)
        self.assertEqual(self.pp.lbtim.ic, 4)

        self.pp.lbtim.ib = 9
        self.assertEqual(self.pp.lbtim.ib, 9)
        self.assertEqual(self.pp.lbtim[1], 9)

    def test_set_lbuser(self):
        self.pp.stash = "m02s12i003"
        self.assertEqual(self.pp.stash, pp.STASH(2, 12, 3))
        self.pp.lbuser[6] = 5
        self.assertEqual(self.pp.stash, pp.STASH(5, 12, 3))
        self.pp.lbuser[3] = 4321
        self.assertEqual(self.pp.stash, pp.STASH(5, 4, 321))

    def test_set_stash(self):
        self.pp.stash = "m02s12i003"
        self.assertEqual(self.pp.stash, pp.STASH(2, 12, 3))

        self.pp.stash = pp.STASH(3, 13, 4)
        self.assertEqual(self.pp.stash, pp.STASH(3, 13, 4))
        self.assertEqual(self.pp.lbuser[3], self.pp.stash.lbuser3())
        self.assertEqual(self.pp.lbuser[6], self.pp.stash.lbuser6())

        with self.assertRaises(ValueError):
            self.pp.stash = (4, 15, 5)

    def test_lbproc_bad_access(self):
        try:
            print(self.pp.lbproc.flag65537)
        except AttributeError:
            pass
        except Exception as err:
            self.fail("Should return a better error: " + str(err))


@tests.skip_data
class TestPPField_GlobalTemperature(IrisPPTest):
    def setUp(self):
        self.original_pp_filepath = tests.get_data_path(
            ("PP", "aPPglob1", "global.pp")
        )
        self.r = list(pp.load(self.original_pp_filepath))

    def test_full_file(self):
        self.check_pp(self.r[0:10], ("PP", "global_test.pp.txt"))

    def test_lbtim_access(self):
        self.assertEqual(self.r[0].lbtim[0], 2)
        self.assertEqual(self.r[0].lbtim.ic, 2)

    def test_t1_t2_access(self):
        self.assertEqual(
            self.r[0].t1.timetuple(),
            cftime.datetime(1994, 12, 1, 0, 0).timetuple(),
        )

    def test_save_single(self):
        temp_filename = iris.util.create_temp_filename(".pp")
        with open(temp_filename, "wb") as temp_fh:
            self.r[0].save(temp_fh)
        self.assertEqual(
            self.file_checksum(temp_filename),
            self.file_checksum(self.original_pp_filepath),
        )
        os.remove(temp_filename)

    def test_save_api(self):
        filepath = self.original_pp_filepath

        f = next(pp.load(filepath))

        temp_filename = iris.util.create_temp_filename(".pp")

        with open(temp_filename, "wb") as temp_fh:
            f.save(temp_fh)
        self.assertEqual(
            self.file_checksum(temp_filename), self.file_checksum(filepath)
        )

        os.remove(temp_filename)


@tests.skip_data
class TestPackedPP(IrisPPTest):
    def test_wgdos(self):
        filepath = tests.get_data_path(
            ("PP", "wgdos_packed", "nae.20100104-06_0001.pp")
        )
        r = pp.load(filepath)

        # Check that the result is a generator and convert to a list so that we
        # can index and get the first one
        self.assertEqual(type(r), GeneratorType)
        r = list(r)

        self.check_pp(r, ("PP", "nae_unpacked.pp.txt"))

        # check that trying to save this field again raises an error
        # (we cannot currently write WGDOS packed fields without mo_pack)
        temp_filename = iris.util.create_temp_filename(".pp")
        with mock.patch("iris.fileformats.pp.mo_pack", None):
            with self.assertRaises(NotImplementedError):
                with open(temp_filename, "wb") as temp_fh:
                    r[0].save(temp_fh)
        os.remove(temp_filename)

    @unittest.skipIf(pp.mo_pack is None, "Requires mo_pack.")
    def test_wgdos_mo_pack(self):
        filepath = tests.get_data_path(
            ("PP", "wgdos_packed", "nae.20100104-06_0001.pp")
        )
        orig_fields = pp.load(filepath)
        with self.temp_filename(".pp") as temp_filename:
            with open(temp_filename, "wb") as fh:
                for field in orig_fields:
                    field.save(fh)
            saved_fields = pp.load(temp_filename)
            for orig_field, saved_field in zip(orig_fields, saved_fields):
                assert_array_equal(orig_field.data, saved_field.data)

    def test_rle(self):
        r = pp.load(tests.get_data_path(("PP", "ocean_rle", "ocean_rle.pp")))

        # Check that the result is a generator and convert to a list so that we
        # can index and get the first one
        self.assertEqual(type(r), GeneratorType)
        r = list(r)

        self.check_pp(r, ("PP", "rle_unpacked.pp.txt"))

        # check that trying to save this field again raises an error
        # (we cannot currently write RLE packed fields)
        with self.temp_filename(".pp") as temp_filename:
            with self.assertRaises(NotImplementedError):
                with open(temp_filename, "wb") as temp_fh:
                    r[0].save(temp_fh)


@tests.skip_data
class TestPPFile(IrisPPTest):
    def test_lots_of_extra_data(self):
        r = pp.load(
            tests.get_data_path(
                ("PP", "cf_processing", "HadCM2_ts_SAT_ann_18602100.b.pp")
            )
        )
        r = list(r)
        self.assertEqual(r[0].lbcode.ix, 13)
        self.assertEqual(r[0].lbcode.iy, 23)
        self.assertEqual(len(r[0].lbcode), 5)
        self.check_pp(r, ("PP", "extra_data_time_series.pp.txt"))


@tests.skip_data
class TestPPFileExtraXData(IrisPPTest):
    def setUp(self):
        self.original_pp_filepath = tests.get_data_path(
            ("PP", "ukV1", "ukVpmslont.pp")
        )
        self.r = list(pp.load(self.original_pp_filepath))[0:5]

    def test_full_file(self):
        self.check_pp(self.r, ("PP", "extra_x_data.pp.txt"))

    def test_save_single(self):
        filepath = tests.get_data_path(
            ("PP", "ukV1", "ukVpmslont_first_field.pp")
        )
        f = next(pp.load(filepath))

        temp_filename = iris.util.create_temp_filename(".pp")
        with open(temp_filename, "wb") as temp_fh:
            f.save(temp_fh)

        s = next(pp.load(temp_filename))

        # force the data to be loaded (this was done for f when save was run)
        s.data
        self._assert_str_same(
            str(s) + "\n", str(f) + "\n", "", type_comparison_name="PP files"
        )

        self.assertEqual(
            self.file_checksum(temp_filename), self.file_checksum(filepath)
        )
        os.remove(temp_filename)


@tests.skip_data
class TestPPFileWithExtraCharacterData(IrisPPTest):
    def setUp(self):
        self.original_pp_filepath = tests.get_data_path(
            ("PP", "globClim1", "dec_subset.pp")
        )
        self.r = pp.load(self.original_pp_filepath)
        self.r_loaded_data = pp.load(self.original_pp_filepath, read_data=True)

        # Check that the result is a generator and convert to a list so that we can index and get the first one
        self.assertEqual(type(self.r), GeneratorType)
        self.r = list(self.r)

        self.assertEqual(type(self.r_loaded_data), GeneratorType)
        self.r_loaded_data = list(self.r_loaded_data)

    def test_extra_field_title(self):
        self.assertEqual(
            self.r[0].field_title,
            "AJHQA Time mean  !C Atmos u compnt of wind after timestep at 9.998 metres !C 01/12/2007 00:00 -> 01/01/2008 00:00",
        )

    def test_full_file(self):
        self.check_pp(self.r[0:10], ("PP", "extra_char_data.pp.txt"))
        self.check_pp(
            self.r_loaded_data[0:10],
            ("PP", "extra_char_data.w_data_loaded.pp.txt"),
        )

    def test_save_single(self):
        filepath = tests.get_data_path(
            ("PP", "model_comp", "dec_first_field.pp")
        )
        f = next(pp.load(filepath))

        temp_filename = iris.util.create_temp_filename(".pp")
        with open(temp_filename, "wb") as temp_fh:
            f.save(temp_fh)

        s = next(pp.load(temp_filename))

        # force the data to be loaded (this was done for f when save was run)
        s.data
        self._assert_str_same(
            str(s) + "\n", str(f) + "\n", "", type_comparison_name="PP files"
        )

        self.assertEqual(
            self.file_checksum(temp_filename), self.file_checksum(filepath)
        )
        os.remove(temp_filename)


class TestSplittableInt(tests.IrisTest):
    def test_3(self):
        t = pp.SplittableInt(3)
        self.assertEqual(t[0], 3)

    def test_grow_str_list(self):
        t = pp.SplittableInt(3)
        t[1] = 3
        self.assertEqual(t[1], 3)

        t[5] = 4

        self.assertEqual(t[5], 4)

        self.assertEqual(int(t), 400033)

        self.assertEqual(t, 400033)
        self.assertNotEqual(t, 33)

        self.assertTrue(t >= 400033)
        self.assertFalse(t >= 400034)

        self.assertTrue(t <= 400033)
        self.assertFalse(t <= 400032)

        self.assertTrue(t > 400032)
        self.assertFalse(t > 400034)

        self.assertTrue(t < 400034)
        self.assertFalse(t < 400032)

    def test_name_mapping(self):
        t = pp.SplittableInt(33214, {"ones": 0, "tens": 1, "hundreds": 2})
        self.assertEqual(t.ones, 4)
        self.assertEqual(t.tens, 1)
        self.assertEqual(t.hundreds, 2)

        t.ones = 9
        t.tens = 4
        t.hundreds = 0

        self.assertEqual(t.ones, 9)
        self.assertEqual(t.tens, 4)
        self.assertEqual(t.hundreds, 0)

    def test_name_mapping_multi_index(self):
        t = pp.SplittableInt(
            33214,
            {
                "weird_number": slice(None, None, 2),
                "last_few": slice(-2, -5, -2),
                "backwards": slice(None, None, -1),
            },
        )
        self.assertEqual(t.weird_number, 324)
        self.assertEqual(t.last_few, 13)
        self.assertRaises(ValueError, setattr, t, "backwards", 1)
        self.assertRaises(ValueError, setattr, t, "last_few", 1)
        self.assertEqual(t.backwards, 41233)
        self.assertEqual(t, 33214)

        t.weird_number = 99
        # notice that this will zero the 5th number

        self.assertEqual(t, 3919)
        t.weird_number = 7899
        self.assertEqual(t, 7083919)
        t.foo = 1

        t = pp.SplittableInt(33214, {"ix": slice(None, 2), "iy": slice(2, 4)})
        self.assertEqual(t.ix, 14)
        self.assertEqual(t.iy, 32)

        t.ix = 21
        self.assertEqual(t, 33221)

        t = pp.SplittableInt(33214, {"ix": slice(-1, 2)})
        self.assertEqual(t.ix, 0)

        t = pp.SplittableInt(4, {"ix": slice(None, 2), "iy": slice(2, 4)})
        self.assertEqual(t.ix, 4)
        self.assertEqual(t.iy, 0)

    def test_33214(self):
        t = pp.SplittableInt(33214)
        self.assertEqual(t[4], 3)
        self.assertEqual(t[3], 3)
        self.assertEqual(t[2], 2)
        self.assertEqual(t[1], 1)
        self.assertEqual(t[0], 4)

        # The rest should be zero
        for i in range(5, 100):
            self.assertEqual(t[i], 0)

    def test_negative_number(self):
        self.assertRaises(ValueError, pp.SplittableInt, -5)
        try:
            _ = pp.SplittableInt(-5)
        except ValueError as err:
            self.assertEqual(
                str(err),
                "Negative numbers not supported with splittable integers object",
            )


class TestSplittableIntEquality(tests.IrisTest):
    def test_not_implemented(self):
        class Terry:
            pass

        sin = pp.SplittableInt(0)
        self.assertIs(sin.__eq__(Terry()), NotImplemented)
        self.assertIs(sin.__ne__(Terry()), NotImplemented)


class TestPPDataProxyEquality(tests.IrisTest):
    def test_not_implemented(self):
        class Terry:
            pass

        pox = pp.PPDataProxy(
            "john",
            "michael",
            "eric",
            "graham",
            "brian",
            "spam",
            "beans",
            "eggs",
        )
        self.assertIs(pox.__eq__(Terry()), NotImplemented)
        self.assertIs(pox.__ne__(Terry()), NotImplemented)


class TestPPFieldEquality(tests.IrisTest):
    def test_not_implemented(self):
        class Terry:
            pass

        pox = pp.PPField3()
        self.assertIs(pox.__eq__(Terry()), NotImplemented)
        self.assertIs(pox.__ne__(Terry()), NotImplemented)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.plot.plot` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.tests.unit.plot import TestGraphicStringCoord
import iris.coords as coords
import iris.coord_systems as ics

if tests.MPL_AVAILABLE:
    import iris.plot as iplt
    import matplotlib.pyplot as plt
    from matplotlib.path import Path
    import cartopy.mpl.geoaxes
    import cartopy.crs as ccrs


@tests.skip_plot
class TestStringCoordPlot(TestGraphicStringCoord):
    def setUp(self):
        super().setUp()
        self.cube = self.cube[0, :]
        self.lat_lon_cube = self.lat_lon_cube[0, :]

    def test_yaxis_labels(self):
        iplt.plot(self.cube, self.cube.coord("str_coord"))
        self.assertBoundsTickLabels("yaxis")

    def test_xaxis_labels(self):
        iplt.plot(self.cube.coord("str_coord"), self.cube)
        self.assertBoundsTickLabels("xaxis")

    def test_yaxis_labels_with_axes(self):
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        iplt.plot(self.cube, self.cube.coord("str_coord"), axes=ax)
        plt.close(fig)
        self.assertBoundsTickLabels("yaxis", ax)

    def test_xaxis_labels_with_axes(self):
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        iplt.plot(self.cube.coord("str_coord"), self.cube, axes=ax)
        plt.close(fig)
        self.assertBoundsTickLabels("xaxis", ax)

    def test_plot_longitude(self):
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        iplt.plot(
            self.lat_lon_cube.coord("longitude"), self.lat_lon_cube, axes=ax
        )
        plt.close(fig)


@tests.skip_plot
class TestTrajectoryWrap(tests.IrisTest):
    """
    Test that a line plot of geographic coordinates wraps around the end of the
    coordinates rather than plotting accross the map.

    """

    def setUp(self):
        plt.figure()
        self.geog_cs = ics.GeogCS(6371229.0)
        self.plate_carree = self.geog_cs.as_cartopy_projection()

    def lon_lat_coords(self, lons, lats, cs=None):
        if cs is None:
            cs = self.geog_cs
        return (
            coords.AuxCoord(
                lons, "longitude", units="degrees", coord_system=cs
            ),
            coords.AuxCoord(
                lats, "latitude", units="degrees", coord_system=cs
            ),
        )

    def assertPathsEqual(self, expected, actual):
        """
        Assert that the given paths are equal once STOP vertices have been
        removed

        """
        expected = expected.cleaned()
        actual = actual.cleaned()
        # Remove Path.STOP vertices
        everts = expected.vertices[np.where(expected.codes != Path.STOP)]
        averts = actual.vertices[np.where(actual.codes != Path.STOP)]
        self.assertArrayAlmostEqual(everts, averts)
        self.assertArrayEqual(expected.codes, actual.codes)

    def check_paths(self, expected_path, expected_path_crs, lines, axes):
        """
        Check that the paths in `lines` match the given expected paths when
        plotted on the given geoaxes

        """

        self.assertEqual(
            1, len(lines), "Expected a single line, got {}".format(len(lines))
        )
        (line,) = lines
        inter_proj_transform = cartopy.mpl.geoaxes.InterProjectionTransform(
            expected_path_crs, axes.projection
        )
        ax_transform = inter_proj_transform + axes.transData

        expected = ax_transform.transform_path(expected_path)
        actual = line.get_transform().transform_path(line.get_path())

        self.assertPathsEqual(expected, actual)

    def test_simple(self):
        lon, lat = self.lon_lat_coords([359, 1], [0, 0])
        expected_path = Path([[-1, 0], [1, 0]], [Path.MOVETO, Path.LINETO])

        lines = iplt.plot(lon, lat)

        self.check_paths(expected_path, self.plate_carree, lines, plt.gca())

    def test_reverse(self):
        lon, lat = self.lon_lat_coords([1, 359], [0, 0])
        expected_path = Path([[1, 0], [-1, 0]], [Path.MOVETO, Path.LINETO])

        lines = iplt.plot(lon, lat)

        self.check_paths(expected_path, self.plate_carree, lines, plt.gca())

    def test_multi(self):
        lon, lat = self.lon_lat_coords([1, 359, 2, 358], [0, 0, 0, 0])
        expected_path = Path(
            [[1, 0], [-1, 0], [2, 0], [-2, 0]],
            [Path.MOVETO, Path.LINETO, Path.LINETO, Path.LINETO],
        )

        lines = iplt.plot(lon, lat)

        self.check_paths(expected_path, self.plate_carree, lines, plt.gca())

    def test_many_wraps(self):
        lon, lat = self.lon_lat_coords(
            [350, 10, 180, 350, 10, 180, 10, 350], [0, 0, 0, 0, 0, 0, 0, 0]
        )
        expected_path = Path(
            [
                [350, 0],
                [370, 0],
                [540, 0],
                [710, 0],
                [730, 0],
                [900, 0],
                [730, 0],
                [710, 0],
            ],
            [
                Path.MOVETO,
                Path.LINETO,
                Path.LINETO,
                Path.LINETO,
                Path.LINETO,
                Path.LINETO,
                Path.LINETO,
                Path.LINETO,
            ],
        )

        lines = iplt.plot(lon, lat)

        self.check_paths(expected_path, self.plate_carree, lines, plt.gca())

    def test_180(self):
        lon, lat = self.lon_lat_coords([179, -179], [0, 0])
        expected_path = Path([[179, 0], [181, 0]], [Path.MOVETO, Path.LINETO])

        lines = iplt.plot(lon, lat)

        self.check_paths(expected_path, self.plate_carree, lines, plt.gca())

    def test_shifted_projection(self):
        lon, lat = self.lon_lat_coords([359, 1], [0, 0])
        expected_path = Path([[-1, 0], [1, 0]], [Path.MOVETO, Path.LINETO])

        shifted_plate_carree = ccrs.PlateCarree(180)

        plt.axes(projection=shifted_plate_carree)
        lines = iplt.plot(lon, lat)

        self.check_paths(expected_path, self.plate_carree, lines, plt.gca())

    def test_shifted_projection_180(self):
        lon, lat = self.lon_lat_coords([179, -179], [0, 0])
        expected_path = Path([[179, 0], [181, 0]], [Path.MOVETO, Path.LINETO])

        shifted_plate_carree = ccrs.PlateCarree(180)

        plt.axes(projection=shifted_plate_carree)
        lines = iplt.plot(lon, lat)

        self.check_paths(expected_path, self.plate_carree, lines, plt.gca())

    def test_long(self):
        lon, lat = self.lon_lat_coords([271, 89], [0, 0])
        expected_path = Path([[-89, 0], [89, 0]], [Path.MOVETO, Path.LINETO])

        lines = iplt.plot(lon, lat)

        self.check_paths(expected_path, self.plate_carree, lines, plt.gca())

    def _test_rotated(
        self,
        grid_north_pole_latitude=90,
        grid_north_pole_longitude=0,
        north_pole_grid_longitude=0,
    ):
        cs = ics.RotatedGeogCS(
            grid_north_pole_latitude,
            grid_north_pole_longitude,
            north_pole_grid_longitude,
        )
        glon = coords.AuxCoord(
            [359, 1], "grid_longitude", units="degrees", coord_system=cs
        )
        glat = coords.AuxCoord(
            [0, 0], "grid_latitude", units="degrees", coord_system=cs
        )
        expected_path = Path([[-1, 0], [1, 0]], [Path.MOVETO, Path.LINETO])

        plt.figure()
        lines = iplt.plot(glon, glat)
        # Matplotlib won't immediately set up the correct transform to allow us
        # to compare paths. Calling set_global(), which calls set_xlim() and
        # set_ylim(), will trigger Matplotlib to set up the transform.
        ax = plt.gca()
        ax.set_global()

        crs = cs.as_cartopy_crs()
        self.check_paths(expected_path, crs, lines, ax)

    def test_rotated_90(self):
        self._test_rotated(north_pole_grid_longitude=90)

    def test_rotated_180(self):
        self._test_rotated(north_pole_grid_longitude=180)

    def test_rotated(self):
        self._test_rotated(
            grid_north_pole_latitude=-30,
            grid_north_pole_longitude=120,
            north_pole_grid_longitude=45,
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integratation tests for the
`iris.aux_factory.OceanSigmaZFactory` class.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import itertools

import numpy as np

from iris._lazy_data import as_lazy_data
from iris.tests.stock import ocean_sigma_z as stock_sample_osz
import iris.util


class Test_sample(tests.IrisTest):
    def setUp(self):
        self.cube = stock_sample_osz()
        # Snapshot result, printed with ...
        #     >>> np.set_printoptions(linewidth=180,
        #            formatter={'float':lambda x:'{:-09.3f}'.format(x)})
        #     >>> print repr(coord.points)
        self.basic_derived_result = np.array(
            [
                [
                    [
                        [-0000.632, -0000.526, -0000.421, -0000.316],
                        [-0000.789, -0000.684, -0000.579, -0000.474],
                        [-0000.947, -0000.842, -0000.737, -0000.632],
                    ],
                    [
                        [-0014.358, -0014.264, -0014.169, -0014.074],
                        [-0014.501, -0014.406, -0014.311, -0014.216],
                        [-0014.643, -0014.548, -0014.453, -0014.358],
                    ],
                    [
                        [-0082.993, -0082.951, -0082.908, -0082.866],
                        [-0083.056, -0083.014, -0082.972, -0082.929],
                        [-0083.119, -0083.077, -0083.035, -0082.993],
                    ],
                    [
                        [-0368.400, -0368.400, -0368.400, -0368.400],
                        [-0368.400, -0368.400, -0368.400, -0368.400],
                        [-0368.400, -0368.400, -0368.400, -0368.400],
                    ],
                    [
                        [-1495.600, -1495.600, -1495.600, -1495.600],
                        [-1495.600, -1495.600, -1495.600, -1495.600],
                        [-1495.600, -1495.600, -1495.600, -1495.600],
                    ],
                ],
                [
                    [
                        [-0000.842, -0000.737, -0000.632, -0000.526],
                        [-0001.000, -0000.895, -0000.789, -0000.684],
                        [-0001.158, -0001.053, -0000.947, -0000.842],
                    ],
                    [
                        [-0014.548, -0014.453, -0014.358, -0014.264],
                        [-0014.690, -0014.595, -0014.501, -0014.406],
                        [-0014.832, -0014.737, -0014.643, -0014.548],
                    ],
                    [
                        [-0083.077, -0083.035, -0082.993, -0082.951],
                        [-0083.140, -0083.098, -0083.056, -0083.014],
                        [-0083.203, -0083.161, -0083.119, -0083.077],
                    ],
                    [
                        [-0368.400, -0368.400, -0368.400, -0368.400],
                        [-0368.400, -0368.400, -0368.400, -0368.400],
                        [-0368.400, -0368.400, -0368.400, -0368.400],
                    ],
                    [
                        [-1495.600, -1495.600, -1495.600, -1495.600],
                        [-1495.600, -1495.600, -1495.600, -1495.600],
                        [-1495.600, -1495.600, -1495.600, -1495.600],
                    ],
                ],
            ]
        )

        self.derived_coord_name = (
            "sea_surface_height_above_reference_ellipsoid"
        )

    def _check_result(self, cube, expected_result=None, **kwargs):
        if expected_result is None:
            expected_result = self.basic_derived_result
        coord = cube.coord(self.derived_coord_name)
        result = coord.points
        self.assertArrayAllClose(result, expected_result, atol=0.005, **kwargs)

    def test_basic(self):
        self._check_result(self.cube)

    def _lazy_testcube(self):
        cube = self.cube
        for dep_name in ("depth", "layer_depth", "ocean_sigma_z_coordinate"):
            coord = cube.coord(dep_name)
            coord.points = as_lazy_data(coord.points, coord.shape)
        return cube

    def test_nonlazy_cube_has_lazy_derived(self):
        # Check same results when key coords are made lazy.
        cube = self.cube
        self.assertEqual(cube.coord("depth").has_lazy_points(), False)
        self.assertEqual(
            cube.coord(self.derived_coord_name).has_lazy_points(), True
        )

    def test_lazy_cube_same_result(self):
        cube = self._lazy_testcube()
        self.assertEqual(cube.coord("depth").has_lazy_points(), True)
        self.assertEqual(
            cube.coord(self.derived_coord_name).has_lazy_points(), True
        )
        self._check_result(cube)

    def test_transpose(self):
        # Check it works with all possible dimension orders.
        for dims_list in itertools.permutations(range(self.cube.ndim)):
            cube = self.cube.copy()
            cube.transpose(dims_list)
            expected = self.basic_derived_result.transpose(dims_list)
            msg = "Unexpected result when cube transposed by {}"
            msg = msg.format(dims_list)
            self._check_result(cube, expected, err_msg=msg)

    def test_lazy_transpose(self):
        # Check lazy calc works with all possible dimension orders.
        for dims_list in itertools.permutations(range(self.cube.ndim)):
            cube = self._lazy_testcube().copy()
            cube.transpose(dims_list)
            expected = self.basic_derived_result.transpose(dims_list)
            msg = "Unexpected result when cube transposed by {}"
            msg = msg.format(dims_list)
            self._check_result(cube, expected, err_msg=msg)

    def test_extra_dims(self):
        # Insert some extra cube dimensions + check it still works.
        cube = self.cube
        cube = iris.util.new_axis(cube)
        cube = iris.util.new_axis(cube)
        cube = iris.util.new_axis(cube)
        # N.B. shape is now (1, 1, 1, t, z, y, x)
        cube.transpose((0, 3, 1, 4, 5, 2, 6))
        # N.B. shape is now (1, t, 1, z, y, 1, x)
        # Should get same original result, as derived dims are the same.
        self._check_result(cube)

    def test_no_sigma(self):
        # Check it still works when 'sigma' is removed.
        # NOTE: the unit test for this does not cover all cases because it
        # doesn't provide a time dimension.

        # Set all sigma points to zero + snapshot the resulting derived points.
        trial_cube = self.cube.copy()
        trial_cube.coord("ocean_sigma_z_coordinate").points[:] = 0.0
        expected = trial_cube.coord(self.derived_coord_name).points

        # Remove sigma altogether + check the result is the same.
        cube = self.cube
        cube.remove_coord("ocean_sigma_z_coordinate")
        self._check_result(cube, expected)

    def test_no_eta(self):
        # Check it still works when 'eta' is removed.
        # NOTE: the unit test for this does not cover all cases because it
        # doesn't provide a time dimension.

        # Set all sigma points to zero + snapshot the resulting derived points.
        trial_cube = self.cube.copy()
        trial_cube.coord("sea_surface_height").points[:] = 0.0
        expected = trial_cube.coord(self.derived_coord_name).points
        # Check this has no variation between the two timepoints.
        self.assertArrayAllClose(expected[0], expected[1])
        # Take first time, as no sigma --> result *has* no time dimension.
        expected = expected[0]

        # Remove eta altogether + check the result is the same.
        cube = self.cube
        cube.remove_coord("sea_surface_height")
        self._check_result(cube, expected)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris._lazy data.lazy_elementwise`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris._lazy_data import as_lazy_data, is_lazy_data

from iris._lazy_data import lazy_elementwise


def _test_elementwise_op(array):
    # Promotes the type of a bool argument, but not a float.
    return array + 1


class Test_lazy_elementwise(tests.IrisTest):
    def test_basic(self):
        concrete_array = np.arange(30).reshape((2, 5, 3))
        lazy_array = as_lazy_data(concrete_array)
        wrapped = lazy_elementwise(lazy_array, _test_elementwise_op)
        self.assertTrue(is_lazy_data(wrapped))
        self.assertArrayAllClose(
            wrapped.compute(), _test_elementwise_op(concrete_array)
        )

    def test_dtype_same(self):
        concrete_array = np.array([3.0], dtype=np.float16)
        lazy_array = as_lazy_data(concrete_array)
        wrapped = lazy_elementwise(lazy_array, _test_elementwise_op)
        self.assertTrue(is_lazy_data(wrapped))
        self.assertEqual(wrapped.dtype, np.float16)
        self.assertEqual(wrapped.compute().dtype, np.float16)

    def test_dtype_change(self):
        concrete_array = np.array([True, False])
        lazy_array = as_lazy_data(concrete_array)
        wrapped = lazy_elementwise(lazy_array, _test_elementwise_op)
        self.assertTrue(is_lazy_data(wrapped))
        self.assertEqual(wrapped.dtype, np.int)
        self.assertEqual(wrapped.compute().dtype, wrapped.dtype)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris.fileformats.pp_load_rules._reduce_points_and_bounds`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.fileformats.pp_load_rules import _reduce_points_and_bounds


class Test(tests.IrisTest):
    def test_scalar(self):
        array = np.array(1)
        dims, result, bounds = _reduce_points_and_bounds(array)
        self.assertArrayEqual(result, array)
        self.assertEqual(dims, None)
        self.assertIsNone(bounds)

    def test_1d_nochange(self):
        array = np.array([1, 2, 3])
        dims, result, _ = _reduce_points_and_bounds(array)
        self.assertArrayEqual(result, array)
        self.assertEqual(dims, (0,))

    def test_1d_collapse(self):
        array = np.array([1, 1, 1])
        dims, result, _ = _reduce_points_and_bounds(array)
        self.assertArrayEqual(result, np.array(1))
        self.assertEqual(dims, None)

    def test_2d_nochange(self):
        array = np.array([[1, 2, 3], [4, 5, 6]])
        dims, result, _ = _reduce_points_and_bounds(array)
        self.assertArrayEqual(result, array)
        self.assertEqual(dims, (0, 1))

    def test_2d_collapse_dim0(self):
        array = np.array([[1, 2, 3], [1, 2, 3]])
        dims, result, _ = _reduce_points_and_bounds(array)
        self.assertArrayEqual(result, np.array([1, 2, 3]))
        self.assertEqual(dims, (1,))

    def test_2d_collapse_dim1(self):
        array = np.array([[1, 1, 1], [2, 2, 2]])
        dims, result, _ = _reduce_points_and_bounds(array)
        self.assertArrayEqual(result, np.array([1, 2]))
        self.assertEqual(dims, (0,))

    def test_2d_collapse_both(self):
        array = np.array([[3, 3, 3], [3, 3, 3]])
        dims, result, _ = _reduce_points_and_bounds(array)
        self.assertArrayEqual(result, np.array(3))
        self.assertEqual(dims, None)

    def test_3d(self):
        array = np.array([[[3, 3, 3], [4, 4, 4]], [[3, 3, 3], [4, 4, 4]]])
        dims, result, _ = _reduce_points_and_bounds(array)
        self.assertArrayEqual(result, np.array([3, 4]))
        self.assertEqual(dims, (1,))

    def test_bounds_collapse(self):
        points = np.array([1, 1, 1])
        bounds = np.array([[0, 2], [0, 2], [0, 2]])
        result_dims, result_pts, result_bds = _reduce_points_and_bounds(
            points, (bounds[..., 0], bounds[..., 1])
        )
        self.assertArrayEqual(result_pts, np.array(1))
        self.assertArrayEqual(result_bds, np.array([0, 2]))
        self.assertEqual(result_dims, None)

    def test_bounds_no_collapse(self):
        points = np.array([1, 2, 3])
        bounds = np.array([[0, 2], [1, 3], [2, 4]])
        result_dims, result_pts, result_bds = _reduce_points_and_bounds(
            points, (bounds[..., 0], bounds[..., 1])
        )
        self.assertArrayEqual(result_pts, points)
        self.assertArrayEqual(result_bds, bounds)
        self.assertEqual(result_dims, (0,))


if __name__ == "__main__":
    tests.main()
